#!/usr/bin/env python
import json
import os

from contextlib import contextmanager
from unittest import TestCase
from unittest.mock import ANY

from rust_decider import Decider
from rust_decider import DeciderException
from rust_decider import DeciderInitException
from rust_decider import FeatureNotFoundException
from rust_decider import ValueTypeMismatchException
from rust_decider import version
from utils import AD_ACCOUNT_ID
from utils import BUSINESS_ID
from utils import CANONICAL_URL
from utils import create_temp_config_file
from utils import ctx_dict
from utils import DEVICE_ID
from utils import event_json
from utils import GEN_EXP_ID
from utils import GEN_EXP_NAME
from utils import GEN_EXP_VERSION
from utils import SUBREDDIT_ID
from utils import USER_ID

TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def setup_decider_class(cfg_path):
    return Decider(cfg_path)


@contextmanager
def test_feature_not_found():
    with create_temp_config_file({}) as f:
        decider = setup_decider_class(f.name)
        yield decider


class TestDecider(TestCase):
    def setUp(self):
        self.valid_ctx_dict = ctx_dict

        variants = [
            {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
            {"range_start": 0.2, "range_end": 0.4, "name": "variant_2"},
            {"range_start": 0.4, "range_end": 0.6, "name": "variant_3"},
            {"range_start": 0.6, "range_end": 0.8, "name": "variant_4"},
            {"range_start": 0.8, "range_end": 1.0, "name": "variant_5"},
        ]

        self.device_id_exp = self.config_for_bucket_val("device_id", variants)

        self.canonical_url_exp = self.config_for_bucket_val("canonical_url", variants)

        self.subreddit_id_exp = self.config_for_bucket_val("subreddit_id", variants)

        self.ad_account_id_exp = self.config_for_bucket_val("ad_account_id", variants)

        self.business_id_exp = self.config_for_bucket_val("business_id", variants)

        self.genexp_0_cfg = {
            "genexp_0": {
                "id": GEN_EXP_ID,
                "name": GEN_EXP_NAME,
                "enabled": True,
                "version": GEN_EXP_VERSION,
                "emit_event": True,
                "type": "range_variant",
                "start_ts": 0,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": variants,
                    "experiment_version": 5,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
        }

        self.additional_2_exp = {
            "exp_0": {
                "id": 3248,
                "name": "exp_0",
                "enabled": True,
                "version": "5",
                "type": "range_variant",
                "emit_event": True,
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [
                        {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
                        {"range_start": 0.2, "range_end": 0.4, "name": "control_2"},
                        {"range_start": 0.4, "range_end": 0.6, "name": "variant_2"},
                        {"range_start": 0.6, "range_end": 0.8, "name": "variant_3"},
                        {"range_start": 0.8, "range_end": 1.0, "name": "variant_4"},
                    ],
                    "experiment_version": 5,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
            "exp_1": {
                "id": 3246,
                "name": "exp_1",
                "enabled": True,
                "version": "5",
                "type": "range_variant",
                "emit_event": True,
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [
                        {"range_start": 0, "range_end": 0, "name": "variant_0"}
                    ],
                    "experiment_version": 5,
                    "shuffle_version": 0,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
        }

        super().setUp()

    def config_for_bucket_val(self, bv, variants):
        genexp_config = {
            "id": 123,
            "name": "genexp_device_id",
            "enabled": True,
            "version": "5",
            "type": "range_variant",
            "start_ts": 0,
            "stop_ts": 2147483648,
            "emit_event": True,
            "experiment": {
                "variants": variants,
                "experiment_version": 5,
                "shuffle_version": 91,
                "bucket_val": None,
                "log_bucketing": False,
            },
        }

        exp_name = f"genexp_{bv}"

        config = {exp_name: {**genexp_config, **{"name": exp_name}}}

        config[exp_name]["experiment"]["bucket_val"] = bv

        return config

    def assert_event_json(
        self,
        ctx,
        event,
        noun,
        variant,
        experiment_name=GEN_EXP_NAME,
        experiment_id=GEN_EXP_ID,
        decision_kind="FracAvail",
        bucket_val=None,
        is_override=False,
    ):
        self.assertEqual(
            dict(event),
            {"decision_kind": decision_kind, "exposure_key": ANY, "json_str": ANY},
        )
        self.assertEqual(
            json.loads(event.json_str),
            event_json(
                ctx,
                noun,
                variant,
                experiment_name,
                experiment_id,
                is_override,
                bucket_val,
            ),
        )

    def test_init(self):
        # handles full cfg.json file
        decider = setup_decider_class(f"{TEST_DIR}/../../cfg.json")
        self.assertEqual(type(decider), Decider)

    def test_init_missing_cfg_file(self):
        with self.assertRaises(DeciderInitException) as e:
            setup_decider_class("foo")
        self.assertEqual(
            str(e.exception),
            "Decider initialization failed: Std io error: No such file or directory (os error 2)",
        )

    def test_init_bad_cfg(self):
        # an experiment's id is string instead of int
        cfg = {
            "exp_0": {
                "id": "3248",
                "name": "exp_0",
                "enabled": True,
                "version": "5",
                "type": "range_variant",
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [],
                    "experiment_version": 5,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            }
        }

        with create_temp_config_file(cfg) as f:
            with self.assertLogs() as captured:
                setup_decider_class(f.name)

            assert any(
                'Partially loaded Decider: 1 feature(s) failed to load: {"exp_0": ParsingError(Error("invalid type: string \\"3248\\", expected u32'
                in x.getMessage()
                for x in captured.records
            )

    def test_init_partially_bad_cfg(self):
        cfg = self.genexp_0_cfg
        invalid_exp = {"some_key": [1, 2, 3]}
        cfg.update(invalid_exp)

        with create_temp_config_file(cfg) as f:
            with self.assertLogs() as captured:
                decider = setup_decider_class(f.name)

                assert any(
                    'Partially loaded Decider: 1 feature(s) failed to load: {"some_key": ParsingError(Error("version field missing or unrecognized"'
                    in x.getMessage()
                    for x in captured.records
                )

                choice = decider.choose("genexp_0", self.valid_ctx_dict)

                self.assertEqual(
                    dict(choice),
                    {
                        "variant": "variant_5",
                        "value": None,
                        "feature_id": 6299,
                        "feature_name": "genexp_0",
                        "feature_version": 5,
                        "events": [
                            f"0::::6299::::genexp_0::::5::::variant_5::::{USER_ID}::::user_id::::0::::2147483648"
                        ],
                        "full_events": ANY,
                    },
                )

                assert len(choice.full_events) == 1
                self.assert_event_json(
                    self.valid_ctx_dict, choice.full_events[0], "user_id", "variant_5"
                )

    def test_choose(self):
        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("genexp_0", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_5",
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [
                        f"0::::6299::::genexp_0::::5::::variant_5::::{USER_ID}::::user_id::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
            assert len(choice.full_events) == 1
            self.assert_event_json(
                self.valid_ctx_dict, choice.full_events[0], "user_id", "variant_5"
            )

    def test_choose_without_variant(self):
        cfg = self.genexp_0_cfg
        variants = [
            {"name": "enabled", "size": 0, "range_end": 0, "range_start": 0},
            {"name": "control_1", "size": 0, "range_end": 0, "range_start": 0},
        ]
        cfg["genexp_0"]["experiment"]["variants"] = variants

        with create_temp_config_file(cfg) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("genexp_0", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": None,
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [],
                    "full_events": [],
                },
            )

    def test_choose_without_ctx(self):
        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider_class(f.name)

            with self.assertRaises(DeciderException) as e:
                decider.choose("genexp_0", None)

            self.assertEqual(
                str(e.exception),
                "Missing `context` param for feature_name: genexp_0",
            )

    def test_choose_bucket_val_device_id(self):
        with create_temp_config_file(self.device_id_exp) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("genexp_device_id", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_4",
                    "value": None,
                    "feature_id": 123,
                    "feature_name": "genexp_device_id",
                    "feature_version": 5,
                    "events": [
                        f"0::::123::::genexp_device_id::::5::::variant_4::::{DEVICE_ID}::::device_id::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
            assert len(choice.full_events) == 1
            self.assert_event_json(
                self.valid_ctx_dict,
                choice.full_events[0],
                "device_id",
                "variant_4",
                "genexp_device_id",
                123,
            )

    def test_choose_bucket_val_device_id_missing_identifier(self):
        with create_temp_config_file(self.device_id_exp) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()
            del ctx["device_id"]

            with self.assertRaises(DeciderException) as e:
                decider.choose("genexp_device_id", ctx)

            self.assertEqual(
                str(e.exception),
                'Missing field "device_id" in context for bucket_val = device_id',
            )

    def test_choose_bucket_val_canonical_url(self):
        with create_temp_config_file(self.canonical_url_exp) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("genexp_canonical_url", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_3",
                    "value": None,
                    "feature_id": 123,
                    "feature_name": "genexp_canonical_url",
                    "feature_version": 5,
                    "events": [
                        f"0::::123::::genexp_canonical_url::::5::::variant_3::::{CANONICAL_URL}::::canonical_url::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
        assert len(choice.full_events) == 1
        self.assert_event_json(
            self.valid_ctx_dict,
            choice.full_events[0],
            "canonical_url",
            "variant_3",
            "genexp_canonical_url",
            123,
        )

    def test_choose_bucket_val_canonical_url_missing_ctx_field(self):
        with create_temp_config_file(self.canonical_url_exp) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()
            del ctx["canonical_url"]

            with self.assertRaises(DeciderException) as e:
                decider.choose("genexp_canonical_url", ctx)
            self.assertEqual(
                str(e.exception),
                'Missing field "canonical_url" in context for bucket_val = canonical_url',
            )

    def test_choose_bucket_val_subreddit_id(self):
        with create_temp_config_file(self.subreddit_id_exp) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("genexp_subreddit_id", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_4",
                    "value": None,
                    "feature_id": 123,
                    "feature_name": "genexp_subreddit_id",
                    "feature_version": 5,
                    "events": [
                        f"0::::123::::genexp_subreddit_id::::5::::variant_4::::{SUBREDDIT_ID}::::subreddit_id::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
            assert len(choice.full_events) == 1
            self.assert_event_json(
                self.valid_ctx_dict,
                choice.full_events[0],
                "subreddit_id",
                "variant_4",
                "genexp_subreddit_id",
                123,
            )

    def test_choose_bucket_val_subreddit_id_missing_ctx_field(self):
        with create_temp_config_file(self.subreddit_id_exp) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()
            del ctx["subreddit_id"]

            with self.assertRaises(DeciderException) as e:
                decider.choose("genexp_subreddit_id", ctx)
            self.assertEqual(
                str(e.exception),
                'Missing field "subreddit_id" in context for bucket_val = subreddit_id',
            )

    def test_choose_bucket_val_ad_account_id(self):
        with create_temp_config_file(self.ad_account_id_exp) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("genexp_ad_account_id", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_2",
                    "value": None,
                    "feature_id": 123,
                    "feature_name": "genexp_ad_account_id",
                    "feature_version": 5,
                    "events": [
                        f"0::::123::::genexp_ad_account_id::::5::::variant_2::::{AD_ACCOUNT_ID}::::ad_account_id::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
            assert len(choice.full_events) == 1
            self.assert_event_json(
                self.valid_ctx_dict,
                choice.full_events[0],
                "ad_account_id",
                "variant_2",
                "genexp_ad_account_id",
                123,
            )

    def test_choose_bucket_val_ad_account_id_missing_ctx_field(self):
        with create_temp_config_file(self.ad_account_id_exp) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()
            del ctx["ad_account_id"]

            with self.assertRaises(DeciderException) as e:
                decider.choose("genexp_ad_account_id", ctx)
            self.assertEqual(
                str(e.exception),
                'Missing field "ad_account_id" in context for bucket_val = ad_account_id',
            )

    def test_choose_bucket_val_business_id(self):
        with create_temp_config_file(self.business_id_exp) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("genexp_business_id", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_3",
                    "value": None,
                    "feature_id": 123,
                    "feature_name": "genexp_business_id",
                    "feature_version": 5,
                    "events": [
                        f"0::::123::::genexp_business_id::::5::::variant_3::::{BUSINESS_ID}::::business_id::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
            assert len(choice.full_events) == 1
            self.assert_event_json(
                self.valid_ctx_dict,
                choice.full_events[0],
                "business_id",
                "variant_3",
                "genexp_business_id",
                123,
            )

    def test_choose_bucket_val_business_id_missing_ctx_field(self):
        with create_temp_config_file(self.business_id_exp) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()
            del ctx["business_id"]

            with self.assertRaises(DeciderException) as e:
                decider.choose("genexp_business_id", ctx)
            self.assertEqual(
                str(e.exception),
                'Missing field "business_id" in context for bucket_val = business_id',
            )

    def test_choose_with_other_fields_for_bucketing(self):
        cfg = self.genexp_0_cfg
        cfg["genexp_0"]["experiment"].update(
            {
                "bucket_val": "foo",
                "variants": [
                    {"range_start": 0.0, "range_end": 1.0, "name": "control_1"},
                ],
            }
        )

        with create_temp_config_file(cfg) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()

            # include bucket_val value in "other_fields"
            ctx.update({"other_fields": {"foo": "bar"}})

            choice = decider.choose("genexp_0", ctx)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "control_1",
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [
                        "0::::6299::::genexp_0::::5::::control_1::::bar::::foo::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
            assert len(choice.full_events) == 1
            self.assert_event_json(
                ctx=ctx,
                event=choice.full_events[0],
                noun="foo",
                variant="control_1",
                bucket_val="bar",
            )

    def test_choose_with_other_fields_for_targeting(self):
        cfg = self.genexp_0_cfg
        cfg["genexp_0"]["experiment"].update(
            {"targeting": {"ALL": [{"EQ": {"field": "foo", "values": ["bar"]}}]}}
        )

        with create_temp_config_file(cfg) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()

            # targeting matches
            ctx.update({"other_fields": {"foo": "bar"}})

            choice = decider.choose("genexp_0", ctx)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_5",
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [
                        f"0::::6299::::genexp_0::::5::::variant_5::::{USER_ID}::::user_id::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
            assert len(choice.full_events) == 1
            self.assert_event_json(
                ctx=ctx,
                event=choice.full_events[0],
                noun="user_id",
                variant="variant_5",
            )

            # targeting doesn't match
            ctx.update({"other_fields": {"foo": "huh"}})

            choice = decider.choose("genexp_0", ctx)
            self.assertEqual(
                dict(choice),
                {
                    "variant": None,
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [],
                    "full_events": [],
                },
            )

    def test_choose_with_other_fields_for_overrides(self):
        cfg = self.genexp_0_cfg
        cfg["genexp_0"]["experiment"].update(
            {
                "bucket_val": "foo",
                "overrides": [
                    {
                        "variant_5": {
                            "ANY": [{"EQ": {"field": "foo", "values": ["bar"]}}]
                        }
                    }
                ],
                "variants": [
                    {"range_start": 0.0, "range_end": 0.0, "name": "control_1"},
                    {"range_start": 0.0, "range_end": 0.0, "name": "variant_5"},
                ],
            }
        )

        with create_temp_config_file(cfg) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()

            # override matches
            ctx.update({"other_fields": {"foo": "bar"}})

            choice = decider.choose("genexp_0", ctx)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_5",
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [
                        "1::::6299::::genexp_0::::5::::variant_5::::bar::::foo::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )
            assert len(choice.full_events) == 1
            self.assert_event_json(
                ctx=ctx,
                event=choice.full_events[0],
                noun="foo",
                variant="variant_5",
                decision_kind="Override",
                bucket_val="bar",
                is_override=True,
            )

            # override doesn't match
            ctx.update({"other_fields": {"foo": "huh"}})

            choice = decider.choose("genexp_0", ctx)
            self.assertEqual(
                dict(choice),
                {
                    "variant": None,
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [],
                    "full_events": [],
                },
            )

    def test_choose_feature_not_found(self):
        with test_feature_not_found() as d:
            with self.assertRaises(FeatureNotFoundException) as e:
                d.choose("any", self.valid_ctx_dict)

            self.assertEqual(
                str(e.exception),
                'Feature "any" not found.',
            )

    def test_choose_with_a_value(self):
        cfg = {
            "dc_bool": {
                "id": 3393,
                "value": True,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "name": "dc_bool",
                "value_type": "Boolean",
                "experiment": {"experiment_version": 2},
            }
        }
        with create_temp_config_file(cfg) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("dc_bool", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": None,
                    "value": True,
                    "feature_id": 3393,
                    "feature_name": "dc_bool",
                    "feature_version": 2,
                    "events": [],
                    "full_events": [],
                },
            )

    def test_measured_rollouts(self):
        configs = json.load(open(f"{TEST_DIR}/../../test.json"))

        cfg = {
            "feature_rollout_100": configs["feature_rollout_100"],
            "measured_rollout_100": configs["measured_rollout_100"],
        }

        with create_temp_config_file(cfg) as f:
            decider = setup_decider_class(f.name)

            choice = decider.choose("feature_rollout_100", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "enabled",
                    "value": None,
                    "feature_id": 9110,
                    "feature_name": "feature_rollout_100",
                    "feature_version": 1,
                    # FR doesn't emit events (due to `emit_event: false``)
                    "events": [],
                    "full_events": [],
                },
            )

            choice = decider.choose("measured_rollout_100", self.valid_ctx_dict)

            self.assertEqual(
                dict(choice),
                {
                    "variant": "enabled",
                    "value": None,
                    "feature_id": 9119,
                    "feature_name": "measured_rollout_100",
                    "feature_version": 1,
                    # MR emits event
                    "events": [
                        f"0::::9119::::measured_rollout_100::::1::::enabled::::{USER_ID}::::user_id::::1522306800::::32533405261"
                    ],
                    "full_events": ANY,
                },
            )

            # MR sets "measured" to True
            mr_feature = decider.get_feature("measured_rollout_100")
            self.assertEqual(mr_feature.measured, True)

    def test_choose_all(self):
        self.genexp_0_cfg.update(self.additional_2_exp)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider_class(f.name)

            decisions = decider.choose_all(self.valid_ctx_dict)

        # all experiments are returned
        # including "exp_1", which has 0% bucketing ("variant": None)
        self.assertEqual(len(decisions), 3)

        # assert genexp_0
        self.assertEqual(
            dict(decisions["genexp_0"]),
            {
                "variant": "variant_5",
                "value": None,
                "feature_version": 5,
                "feature_id": 6299,
                "feature_name": "genexp_0",
                "events": [
                    f"0::::6299::::genexp_0::::5::::variant_5::::{USER_ID}::::user_id::::0::::2147483648"
                ],
                "full_events": ANY,
            },
        )
        assert len(decisions["genexp_0"].full_events) == 1
        self.assert_event_json(
            ctx=self.valid_ctx_dict,
            event=decisions["genexp_0"].full_events[0],
            noun="user_id",
            variant="variant_5",
        )

        # assert exp_0
        self.assertEqual(
            dict(decisions["exp_0"]),
            {
                "variant": "variant_3",
                "value": None,
                "feature_version": 5,
                "feature_id": 3248,
                "feature_name": "exp_0",
                "events": [
                    f"0::::3248::::exp_0::::5::::variant_3::::{USER_ID}::::user_id::::37173982::::2147483648"
                ],
                "full_events": ANY,
            },
        )

        assert len(decisions["exp_0"].full_events) == 1
        self.assert_event_json(
            ctx=self.valid_ctx_dict,
            event=decisions["exp_0"].full_events[0],
            noun="user_id",
            variant="variant_3",
            experiment_name="exp_0",
            experiment_id=3248,
        )

        # assert exp_1 is included even though "variant" is `None`
        self.assertEqual(
            dict(decisions["exp_1"]),
            {
                "variant": None,
                "value": None,
                "feature_version": 5,
                "feature_id": 3246,
                "feature_name": "exp_1",
                "events": [],
                "full_events": [],
            },
        )

    def test_choose_all_without_ctx(self):
        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider_class(f.name)

            with self.assertRaises(DeciderException) as e:
                decider.choose_all(None)

        self.assertEqual(
            str(e.exception),
            "Missing `context` param",
        )

    def test_choose_all_with_single_identifier_type(self):
        self.genexp_0_cfg.update(self.additional_2_exp)
        self.genexp_0_cfg.update(self.canonical_url_exp)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider_class(f.name)

            decisions = decider.choose_all(
                context=self.valid_ctx_dict, bucketing_field_filter="canonical_url"
            )

        # assert canonical_url_exp
        self.assertEqual(len(decisions), 1)
        self.assertEqual(
            dict(decisions["genexp_canonical_url"]),
            {
                "variant": "variant_3",
                "value": None,
                "feature_version": 5,
                "feature_id": 123,
                "feature_name": "genexp_canonical_url",
                "events": [
                    f"0::::123::::genexp_canonical_url::::5::::variant_3::::{CANONICAL_URL}::::canonical_url::::0::::2147483648"
                ],
                "full_events": ANY,
            },
        )
        assert len(decisions["genexp_canonical_url"].full_events) == 1
        self.assert_event_json(
            ctx=self.valid_ctx_dict,
            event=decisions["genexp_canonical_url"].full_events[0],
            noun="canonical_url",
            variant="variant_3",
            experiment_name="genexp_canonical_url",
            experiment_id=123,
        )

    def test_choose_all_bucket_val_device_id_missing_identifier(self):
        # 3 exp total
        # 1 of which is a device_id exp
        self.device_id_exp.update(self.additional_2_exp)

        with create_temp_config_file(self.device_id_exp) as f:
            decider = setup_decider_class(f.name)
            ctx = self.valid_ctx_dict.copy()
            del ctx["device_id"]

            decisions = decider.choose_all(ctx)

        # device_id exp is excluded from return value
        # since it failed bucketing due to missing "device_id" in context
        self.assertEqual(len(decisions), 2)

        # assert exp_0
        self.assertEqual(
            dict(decisions["exp_0"]),
            {
                "variant": "variant_3",
                "value": None,
                "feature_version": 5,
                "feature_id": 3248,
                "feature_name": "exp_0",
                "events": [
                    f"0::::3248::::exp_0::::5::::variant_3::::{USER_ID}::::user_id::::37173982::::2147483648"
                ],
                "full_events": ANY,
            },
        )

        assert len(decisions["exp_0"].full_events) == 1
        self.assert_event_json(
            ctx=ctx,
            event=decisions["exp_0"].full_events[0],
            noun="user_id",
            variant="variant_3",
            experiment_name="exp_0",
            experiment_id=3248,
        )

        # assert exp_1 is included even though "variant" is `None`
        self.assertEqual(
            dict(decisions["exp_1"]),
            {
                "variant": None,
                "value": None,
                "feature_version": 5,
                "feature_id": 3246,
                "feature_name": "exp_1",
                "events": [],
                "full_events": [],
            },
        )

    def test_get_feature(self):
        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider_class(f.name)

            feature = decider.get_feature("genexp_0")

            cfg = self.genexp_0_cfg["genexp_0"]
            self.assertEqual(feature.id, cfg["id"])
            self.assertEqual(feature.name, cfg["name"])
            self.assertEqual(feature.version, cfg["experiment"]["experiment_version"])
            self.assertEqual(feature.bucket_val, cfg["experiment"]["bucket_val"])
            self.assertEqual(feature.start_ts, cfg["start_ts"])
            self.assertEqual(feature.stop_ts, cfg["stop_ts"])
            self.assertEqual(feature.emit_event, True)
            self.assertEqual(feature.measured, False)

            self.assertEqual(
                dict(feature),
                {
                    "id": cfg["id"],
                    "name": cfg["name"],
                    "version": cfg["experiment"]["experiment_version"],
                    "bucket_val": cfg["experiment"]["bucket_val"],
                    "start_ts": cfg["start_ts"],
                    "stop_ts": cfg["stop_ts"],
                    "emit_event": True,
                    "measured": False,
                },
            )

    def test_get_feature_does_not_exist(self):
        with test_feature_not_found() as d:
            with self.assertRaises(FeatureNotFoundException) as e:
                d.get_feature("any")

            self.assertEqual(
                str(e.exception),
                'Feature "any" not found.',
            )

    def test_version(self):
        release_version = json.load(
            open(f"{TEST_DIR}/../../.release-please-manifest.json")
        )["."]
        v = version()
        version_check = v == "0.0.1-dev" or v == release_version
        self.assertEqual(version_check, True)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider_class(f.name)
            decider_version_check = (
                decider._pkg_version == release_version
                or decider._pkg_version == "0.0.1-dev"
            )
            self.assertEqual(decider_version_check, True)


class TestDeciderGetDynamicConfig(TestCase):
    def setUp(self):
        super().setUp()

        self.valid_ctx_dict = ctx_dict
        self.dc_base_config = {
            "dc_1": {
                "id": 1,
                "name": "dc_1",
                "enabled": True,
                "version": "1",
                "type": "dynamic_config",
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "owner": "test_owner",
                "experiment": {
                    "experiment_version": 1,
                },
            }
        }

    def test_get_bool(self):
        bool_val = True
        self.dc_base_config["dc_1"].update({"value_type": "Boolean", "value": bool_val})

        with create_temp_config_file(self.dc_base_config) as f:
            decider = setup_decider_class(f.name)
            res = decider.get_bool("dc_1", {})
            self.assertEqual(res, bool_val)

            # test type mismatch
            with self.assertRaises(ValueTypeMismatchException) as e:
                decider.get_float("dc_1", {})

            self.assertEqual(
                str(e.exception),
                'Feature "dc_1" not of float type.',
            )

    def test_get_feature_does_not_exist(self):
        with test_feature_not_found() as d:
            with self.assertRaises(FeatureNotFoundException) as e:
                d.get_bool("any", {})

            self.assertEqual(
                str(e.exception),
                'Feature "any" not found.',
            )

    def test_get_bool_missing_ctx(self):
        bool_val = True
        self.dc_base_config["dc_1"].update({"value_type": "Boolean", "value": bool_val})

        with create_temp_config_file(self.dc_base_config) as f:
            decider = setup_decider_class(f.name)
            with self.assertRaises(DeciderException) as e:
                decider.get_bool("dc_1", None)

            self.assertEqual(
                str(e.exception),
                "Missing `context` param for feature_name: dc_1",
            )

    def test_get_int(self):
        int_val = 7
        self.dc_base_config["dc_1"].update({"value_type": "Integer", "value": int_val})

        with create_temp_config_file(self.dc_base_config) as f:
            decider = setup_decider_class(f.name)

            res = decider.get_int("dc_1", {})
            self.assertEqual(res, int_val)

            # test type mismatch
            with self.assertRaises(ValueTypeMismatchException) as e:
                decider.get_bool("dc_1", {})

            self.assertEqual(
                str(e.exception),
                'Feature "dc_1" not of boolean type.',
            )

    def test_get_float(self):
        float_val = 4.20
        self.dc_base_config["dc_1"].update({"value_type": "Float", "value": float_val})

        with create_temp_config_file(self.dc_base_config) as f:
            decider = setup_decider_class(f.name)

            res = decider.get_float("dc_1", {})
            self.assertEqual(res, float_val)

            # test type mismatch
            with self.assertRaises(ValueTypeMismatchException) as e:
                decider.get_map("dc_1", {})

            self.assertEqual(
                str(e.exception),
                'Feature "dc_1" not of map type.',
            )

    def test_get_string(self):
        string_val = "helloworld!"
        self.dc_base_config["dc_1"].update({"value_type": "Text", "value": string_val})

        with create_temp_config_file(self.dc_base_config) as f:
            decider = setup_decider_class(f.name)

            res = decider.get_string("dc_1", {})
            self.assertEqual(res, string_val)

            # test type mismatch
            with self.assertRaises(ValueTypeMismatchException) as e:
                decider.get_int("dc_1", {})

            self.assertEqual(
                str(e.exception),
                'Feature "dc_1" not of integer type.',
            )

    def test_get_map(self):
        map_val = {
            "v": {"nested_map": {"w": False, "x": 1, "y": "some_string", "z": 3.0}},
            "w": False,
            "x": 1,
            "y": "some_string",
            "z": 3.0,
        }
        self.dc_base_config["dc_1"].update({"value_type": "Map", "value": map_val})

        with create_temp_config_file(self.dc_base_config) as f:
            decider = setup_decider_class(f.name)

            res = decider.get_map("dc_1", {})
            self.assertEqual(res, map_val)

            # test type mismatch
            with self.assertRaises(ValueTypeMismatchException) as e:
                decider.get_string("dc_1", {})

            self.assertEqual(
                str(e.exception),
                'Feature "dc_1" not of string type.',
            )

    def test_get_all_values(self):
        base_cfg = self.dc_base_config["dc_1"].copy()

        bool_val = True
        cfg_bool = {"dc_bool": base_cfg.copy()}
        cfg_bool["dc_bool"].update(
            {"name": "dc_bool", "value": bool_val, "value_type": "Boolean"}
        )

        cfg_missing_bool = {}
        cfg_missing_bool["dc_missing_bool"] = cfg_bool["dc_bool"].copy()
        cfg_missing_bool["dc_missing_bool"].update(
            {"value": None, "name": "dc_missing_bool"}
        )

        int_val = 99
        cfg_int = {"dc_int": base_cfg.copy()}
        cfg_int["dc_int"].update(
            {"name": "dc_int", "value": int_val, "value_type": "Integer"}
        )

        cfg_missing_int = {}
        cfg_missing_int["dc_missing_int"] = cfg_int["dc_int"].copy()
        cfg_missing_int["dc_missing_int"].update(
            {"value": None, "name": "dc_missing_int"}
        )

        float_val = 3.2
        cfg_float = {"dc_float": base_cfg.copy()}
        cfg_float["dc_float"].update(
            {"name": "dc_float", "value": float_val, "value_type": "Float"}
        )

        cfg_missing_float = {}
        cfg_missing_float["dc_missing_float"] = cfg_float["dc_float"].copy()
        cfg_missing_float["dc_missing_float"].update(
            {"value": None, "name": "dc_missing_float"}
        )

        string_val = "some_string"
        cfg_string = {"dc_string": base_cfg.copy()}
        cfg_string["dc_string"].update(
            {"name": "dc_string", "value": string_val, "value_type": "String"}
        )
        cfg_text = {"dc_text": base_cfg.copy()}
        cfg_text["dc_text"].update(
            {"name": "dc_text", "value": string_val, "value_type": "Text"}
        )

        cfg_missing_string = {}
        cfg_missing_string["dc_missing_string"] = cfg_string["dc_string"].copy()
        cfg_missing_string["dc_missing_string"].update(
            {"value": None, "name": "dc_missing_string"}
        )

        cfg_missing_text = {}
        cfg_missing_text["dc_missing_text"] = cfg_text["dc_text"].copy()
        cfg_missing_text["dc_missing_text"].update(
            {"value": None, "name": "dc_missing_text"}
        )

        map_val = {
            "v": {"nested_map": {"w": True, "x": 1, "y": "some_string", "z": 3.0}},
            "w": False,
            "x": 1,
            "y": "some_string",
            "z": 3.0,
        }
        cfg_map = {"dc_map": base_cfg.copy()}
        cfg_map["dc_map"].update(
            {"name": "dc_map", "value": map_val, "value_type": "Map"}
        )

        cfg_missing_map = {}
        cfg_missing_map["dc_missing_map"] = cfg_map["dc_map"].copy()
        cfg_missing_map["dc_missing_map"].update(
            {"value": None, "name": "dc_missing_map"}
        )

        missing_value_type_cfg = {
            "dc_missing_value_type": {
                "id": 3393,
                "value": False,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "name": "dc_missing_value_type",
                "experiment": {"experiment_version": 2},
            }
        }

        experiments_cfg = {
            "genexp_0": {
                "id": 6299,
                "name": "genexp_0",
                "enabled": True,
                "version": "5",
                "emit_event": True,
                "type": "range_variant",
                "start_ts": 0,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [
                        {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
                        {"range_start": 0.2, "range_end": 0.4, "name": "variant_2"},
                        {"range_start": 0.4, "range_end": 0.6, "name": "variant_3"},
                        {"range_start": 0.6, "range_end": 0.8, "name": "variant_4"},
                        {"range_start": 0.8, "range_end": 1.0, "name": "variant_5"},
                    ],
                    "experiment_version": 5,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
            "exp_0": {
                "id": 3248,
                "name": "exp_0",
                "enabled": True,
                "version": "5",
                "type": "range_variant",
                "emit_event": True,
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [
                        {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
                        {"range_start": 0.2, "range_end": 0.4, "name": "control_2"},
                        {"range_start": 0.4, "range_end": 0.6, "name": "variant_2"},
                        {"range_start": 0.6, "range_end": 0.8, "name": "variant_3"},
                        {"range_start": 0.8, "range_end": 1.0, "name": "variant_4"},
                    ],
                    "experiment_version": 2,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
            "exp_1": {
                "id": 3246,
                "name": "exp_1",
                "enabled": True,
                "version": "5",
                "type": "range_variant",
                "emit_event": True,
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [
                        {"range_start": 0, "range_end": 0, "name": "variant_0"}
                    ],
                    "experiment_version": 2,
                    "shuffle_version": 0,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
        }
        experiments_cfg.update(cfg_bool)
        experiments_cfg.update(cfg_int)
        experiments_cfg.update(cfg_float)
        experiments_cfg.update(cfg_string)
        experiments_cfg.update(cfg_text)
        experiments_cfg.update(cfg_map)

        # should be set to default values
        experiments_cfg.update(cfg_missing_bool)
        experiments_cfg.update(cfg_missing_int)
        experiments_cfg.update(cfg_missing_float)
        experiments_cfg.update(cfg_missing_string)
        experiments_cfg.update(cfg_missing_text)
        experiments_cfg.update(cfg_missing_map)

        # missing "value_type" field
        experiments_cfg.update(missing_value_type_cfg)

        with create_temp_config_file(experiments_cfg) as f:
            decider = setup_decider_class(f.name)

            configs = decider.all_values({})

            # 6 correct DCs, 6 DCs w/ values set to respective defaults
            # 1 `missing_value_type_cfg` which sets "type" to empty string
            # (3 regular experiments are excluded)
            self.assertEqual(len(configs), 13)

            # test values get set
            self.assertEqual(configs["dc_bool"], bool_val)

            self.assertEqual(configs["dc_int"], int_val)

            self.assertEqual(configs["dc_float"], float_val)

            self.assertEqual(configs["dc_string"], string_val)

            self.assertEqual(configs["dc_text"], string_val)

            self.assertEqual(configs["dc_map"], map_val)

            # test default values
            self.assertEqual(configs["dc_missing_bool"], False)

            self.assertEqual(configs["dc_missing_int"], 0)

            self.assertEqual(configs["dc_missing_float"], 0.0)

            self.assertEqual(configs["dc_missing_string"], "")

            self.assertEqual(configs["dc_missing_text"], "")

            self.assertEqual(configs["dc_missing_map"], {})

            # set "type" to empty string if "value_type" is missing on cfg
            self.assertEqual(configs["dc_missing_value_type"], False)



import json
import random
import tempfile

from contextlib import contextmanager
from typing import Optional
from unittest.mock import ANY

USER_ID = "795244"
AD_ACCOUNT_ID = "t2_4321"
APP_NAME = "ios"
APP_VERSION = "0.0.0.0"
AUTH_CLIENT_ID = "token"
BUILD_NUMBER = 1234
BUSINESS_ID = "t_some"
CANONICAL_URL = "www.reddit.com"
BASE_URL = "https://www.reddit.com?test=true"
USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36"
REFERRER_URL = "https://www.google.com"
COOKIE_CREATED_TIMESTAMP = 1648859753
COUNTRY_CODE = "UA"
DEVICE_ID = "d42b90e7-aae3-4ac5-b137-042de165ecf6"
IS_EMPLOYEE = True
IS_LOGGED_IN = True
EVENT_FIELDS = {
    "user_id": USER_ID,
    "logged_in": IS_LOGGED_IN,
    "cookie_created_timestamp": COOKIE_CREATED_TIMESTAMP,
}
LOID_CREATED_TIMESTAMP = 123456
LOCALE_CODE = "us_en"
ORIGIN_SERVICE = "origin"
SUBREDDIT_ID = "t5_asdf"

EXPERIMENT_OWNER = "test"
GEN_EXP_ID = 6299
GEN_EXP_NAME = "genexp_0"
GEN_EXP_VERSION = "5"

ctx_dict = {
    "user_id": USER_ID,
    "device_id": DEVICE_ID,
    "canonical_url": CANONICAL_URL,
    "base_url": BASE_URL,
    "user_agent": USER_AGENT,
    "referrer_url": REFERRER_URL,
    "subreddit_id": SUBREDDIT_ID,
    "ad_account_id": AD_ACCOUNT_ID,
    "business_id": BUSINESS_ID,
    "locale": LOCALE_CODE,
    "user_is_employee": IS_EMPLOYEE,
    "logged_in": IS_LOGGED_IN,
    "app_name": APP_NAME,
    "build_number": BUILD_NUMBER,
    "country_code": COUNTRY_CODE,
    "origin_service": "oss",
    "oauth_client_id": "test",
    "cookie_created_timestamp": COOKIE_CREATED_TIMESTAMP,
}

# TODO: make these real country codes.  Not necessary, but nicer.
# eventually get from reddit-service-admin/admin/lib/geolocations.py
LOCALES = ["us", "gb", "es"]
APP_NAMES = [
    "app_1",
    "app_2",
    "app_3",
    "app_4",
    "app_5",
]  # TODO: figure out if this is enough.


@contextmanager
def create_temp_config_file(contents):
    with tempfile.NamedTemporaryFile() as f:
        f.write(json.dumps(contents).encode())
        f.seek(0)
        yield f


def make_request_context_map(h={}):
    return {
        "user_id": str(h.get("user_id", random.choice(range(1000)))),
        "locale": h.get("locale", random.choice(LOCALES + [None])),
        "device_id": str(h.get("device_id", 10000 + random.choice(range(1000)))),
        "country_code": h.get("locale", random.choice(LOCALES + [None])),
        "origin_service": h.get("origin_service"),  # do I care about this one?
        "user_is_employee": h.get(
            "user_is_employee", random.choice([True, False, None])
        ),
        "logged_in": h.get("logged_in", random.choice([True, False, None])),
        "app_name": h.get("app_name", random.choice(APP_NAMES + [None])),
        "build_number": int(h.get("build_number", 1000 + random.choice(range(1000)))),
    }


def make_experiment(n, h={}):
    version = int(h.get("version", random.choice(range(10000))))
    variants = h.get(
        "variants", h.get("experiment", {}).get("variants", make_variants())
    )
    shuffle_version = h.get(
        "shuffle_version",
        h.get("experiment", {}).get("shuffle_version", random.choice(range(100))),
    )
    return {
        "id": int(h.get("id", random.choice(range(10000)))),
        "name": str(h.get("name", "genexp_" + str(n))),
        "enabled": True,  # TODO: make this false sometimes?  lowpri
        "owner": "test",
        "version": str(version),
        "type": "range_variant",
        "start_ts": 0,  # we're not interested in testing start/stop
        "stop_ts": 2147483648,
        "experiment": {
            "variants": variants,
            "experiment_version": int(version),
            "shuffle_version": int(shuffle_version),
            "bucket_val": "user_id",  # TODO: make this handle device_id, etc.
            "log_bucketing": False,
            # "overrides": {}, # TODO: build this.
            # "targeting": targeting_tree(),
        },
    }


def make_overrides(variants):
    names = [v["name"] for v in variants]
    return {n: targeting_tree() for n in names}


def targeting_tree():
    """Generate a random targeting tree."""
    return {"EQ": {"field": "user_id", "values": ["3", "5", "7"]}}


def make_variants(h={}):
    return h or [  # TODO: actually generate variantsets.  Lowpri?
        {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
        {"range_start": 0.2, "range_end": 0.4, "name": "variant_2"},
        {"range_start": 0.4, "range_end": 0.6, "name": "variant_3"},
        {"range_start": 0.6, "range_end": 0.8, "name": "variant_4"},
    ]


def event_json(
    ctx: dict,
    noun: str,
    variant: str,
    experiment_name: Optional[str],
    experiment_id: Optional[int],
    is_override: Optional[bool],
    bucket_val: Optional[str] = None,
):
    event_thrift_json = {
        "1": {"str": "experiment"},
        "2": {"str": "expose"},
        "3": {"str": noun},
        "5": {"i64": ANY},
        "6": {"str": ANY},
        "8": {"str": ANY},
        "107": {
            "rec": {
                "2": {"str": APP_NAME},
                "4": {"i32": BUILD_NUMBER},
                "6": {"str": LOCALE_CODE},
            }
        },
        "109": {
            "rec": {
                "1": {"str": USER_AGENT},
                "3": {"str": BASE_URL},
            }
        },
        "110": {"rec": {"2": {"str": REFERRER_URL}}},
        "112": {
            "rec": {
                "1": {"str": USER_ID},
                "3": {"tf": int(IS_LOGGED_IN)},
                "4": {"i64": COOKIE_CREATED_TIMESTAMP},
                "16": {"tf": int(IS_EMPLOYEE)},
            }
        },
        "129": {
            "rec": {
                "1": {"i64": experiment_id},
                "2": {"str": experiment_name},
                "4": {"str": variant},
                "5": {"i64": ANY},
                "6": {"i64": ANY},
                "7": {"str": noun},
                "8": {"str": GEN_EXP_VERSION},
                "9": {"str": bucket_val or ctx_dict[noun]},
                "10": {"tf": int(is_override)},
            }
        },
        "500": {"rec": {"1": {"str": COUNTRY_CODE}}},
    }

    if "device_id" in ctx:
        event_thrift_json["108"] = {"rec": {"2": {"str": DEVICE_ID}}}

    if "canonical_url" in ctx:
        event_thrift_json["109"] = {
            "rec": {"17": {"str": CANONICAL_URL}, **event_thrift_json["109"]["rec"]}
        }

    if "subreddit_id" in ctx:
        event_thrift_json["114"] = {"rec": {"1": {"str": SUBREDDIT_ID}}}

    return event_thrift_json



import logging

from typing import Any
from typing import Callable
from typing import Dict
from typing import ItemsView
from typing import Iterator
from typing import KeysView
from typing import List
from typing import Mapping
from typing import Optional
from typing import Union
from typing import ValuesView

from .events import Exposer
from .prometheus_metrics import decider_client_counter
from .rust_decider import Decider as PyDecider
from .rust_decider import DeciderException
from .rust_decider import DeciderInitException
from .rust_decider import FeatureNotFoundException
from .rust_decider import init  # noqa: F401
from .rust_decider import make_ctx
from .rust_decider import PartialLoadException
from .rust_decider import ValueTypeMismatchException
from .rust_decider import version

logger = logging.getLogger(__name__)

DEFAULT_DECISIONMAKERS = (
    "darkmode overrides targeting holdout mutex_group fractional_availability value"
)

JsonValue = Union[
    None, int, float, str, bool, List["JsonValue"], Mapping[str, "JsonValue"]
]


class BaseMapping(Mapping[str, Any]):
    def keys(self) -> KeysView[str]:
        return self.__dict__.keys()

    def items(self) -> ItemsView[str, Any]:
        return self.__dict__.items()

    def values(self) -> ValuesView[Any]:
        return self.__dict__.values()

    def get(self, key: str, default=None) -> Optional[Any]:
        return self.__dict__.get(key, default)

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def __getitem__(self, key: str) -> Any:
        return self.__dict__[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.__dict__)

    def __len__(self) -> int:
        return len(self.__dict__)


class Decision(BaseMapping):
    def __init__(
        self,
        variant: Optional[str],
        value: Optional[JsonValue],
        feature_id: int,
        feature_name: str,
        feature_version: int,
        events: List[str],
        full_events: List[str],
    ):
        self.variant = variant
        self.value = value
        self.feature_id = feature_id
        self.feature_name = feature_name
        self.feature_version = feature_version
        self.events = events
        self.full_events = full_events


class Feature(BaseMapping):
    def __init__(
        self,
        id: int,
        name: str,
        version: int,
        bucket_val: str,
        start_ts: int,
        stop_ts: int,
        emit_event: bool,
        measured: bool,
        owner: str = None,
    ):
        self.id = id
        self.name = name
        self.version = version
        self.bucket_val = bucket_val
        self.start_ts = start_ts
        self.stop_ts = stop_ts
        self.emit_event = emit_event
        self.measured = measured


class Event(BaseMapping):
    def __init__(
        self,
        decision_kind: str,
        exposure_key: str,
        json_str: str,
    ):
        self.decision_kind = decision_kind
        self.exposure_key = exposure_key
        self.json_str = json_str  # thrift encoded v2 event schema json string


class Decider:
    """Decider class wraps python bindings for Rust decider crate."""

    def __init__(self, path: str, exposer: Optional[Exposer] = None):
        """Initialize a :code:`Decider` class using config file at specified :code:`path`.

        Takes an optional :code:`Exposer` class to emit expose events directly
        (enabled via `expose` bool param in bucketing API).

        Raises :code:`PartialLoadException` if an error occurs when initializing :code:`Decider`
        and any of the features in the config file are malformed.

        Raises general :code:`DeciderInitException` if an error occurs when initializing :code:`Decider`
        (:code:`PartialLoadException` is a sub-exception of :code:`DeciderInitException`).
        """
        self._pkg_version = version()
        self._exposer = exposer

        try:
            self._decider = PyDecider(path)
        except PartialLoadException as e:
            decider_client_counter.labels(
                operation="init",
                success="false",
                error_type="partial_init_exception",
                pkg_version=self._pkg_version,
            ).inc()

            # log errors of misconfigured features
            logger.error(f"{e.args[0]}: {e.args[2]}")

            # set _decider to partially initialized PyDecider instance
            self._decider = e.args[1]
            return
        except DeciderInitException as e:
            decider_client_counter.labels(
                operation="init",
                success="false",
                error_type="init_exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e

        decider_client_counter.labels(
            operation="init",
            success="true",
            error_type="",
            pkg_version=self._pkg_version,
        ).inc()

    def choose(
        self,
        feature_name: str,
        context: Mapping[str, JsonValue],
        expose: Optional[bool] = False,
        expose_holdout: Optional[bool] = False,
    ) -> Decision:
        """Return a :code:`Decision` for a given :code:`feature_name` and :code:`context`.

        If dict() is called on the Decision instance, the following fields are returned:
            .. code-block:: python
            {
                "variant": Optional[str],
                "value": Optional[JsonValue],
                "feature_id": int,
                "feature_name": str,
                "feature_version": int,
                "events": [str]
            }

        Raises :code:`DeciderException` if an error occurs when fetching the Decision.

        :param feature_name: Name of feature you want a variant or value for.

        :param context: A required :code:`Mapping[str, JsonValue]` of context fields used for
            feature overrides, targeting, & bucketing.

        :param expose: Emit expose events via instance's :code:`self._exposer` if set to True.

        :param expose_holdout: Emit expose events for Holdout Groups via instance's :code:`self._exposer` if set to True,
            whih can be combined with :code:`expose=False` to only emit Holdout exposures.

        :return: A :code:`Decision` instance if feature is active/exists,
            raises a :code:`FeatureNotFoundException`
            (a sub-exception of :code:`DeciderException`) otherwise.
        """
        if context is None:
            decider_client_counter.labels(
                operation="choose",
                success="false",
                error_type="missing_context",
                pkg_version=self._pkg_version,
            ).inc()
            raise DeciderException(
                f"Missing `context` param for feature_name: {feature_name}"
            )

        ctx = make_ctx(context)
        ctx_err = ctx.err()
        if ctx_err:
            decider_client_counter.labels(
                operation="choose",
                success="false",
                error_type="invalid_context",
                pkg_version=self._pkg_version,
            ).inc()
            raise DeciderException(
                f"Encountered error from rust_decider.make_ctx() for context: {ctx_err}"
            )

        try:
            decision = self._decider.choose(feature_name, ctx)
        except FeatureNotFoundException as e:
            decider_client_counter.labels(
                operation="choose",
                success="false",
                error_type="feature_not_found",
                pkg_version=self._pkg_version,
            ).inc()
            raise e
        except DeciderException as e:
            decider_client_counter.labels(
                operation="choose",
                success="false",
                error_type="decider_exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e
        except Exception as e:
            decider_client_counter.labels(
                operation="choose",
                success="false",
                error_type="exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e

        decider_client_counter.labels(
            operation="choose",
            success="true",
            error_type="",
            pkg_version=self._pkg_version,
        ).inc()

        if (expose or expose_holdout) and self._exposer:
            for event in decision.events:
                if expose or (expose_holdout and event.decision_kind == "Holdout"):
                    try:
                        self._exposer.expose_event(event.json_str)
                    except Exception as e:
                        decider_client_counter.labels(
                            operation="choose",
                            success="false",
                            error_type="expose",
                            pkg_version=self._pkg_version,
                        ).inc()
                        raise e

        return Decision(
            variant=decision.variant_name,
            feature_id=decision.feature_id,
            feature_name=decision.feature_name,
            feature_version=decision.feature_version,
            value=decision.value,
            events=decision.event_data,  # legacy :::: delimited string format
            full_events=self._map_events(decision.events),
        )

    def _map_events(self, events) -> List[Event]:
        return [
            Event(
                decision_kind=e.decision_kind,
                exposure_key=e.exposure_key,
                json_str=e.json_str,
            )
            for e in events
        ]

    def choose_all(
        self,
        context: Mapping[str, JsonValue],
        bucketing_field_filter: Optional[str] = None,
    ) -> Dict[str, Decision]:
        """Return a :code:`Dict[str, Decision]` for all active features...

        with the key as the feature name (not including features
        of :code:`"type": "dynamic_config"`) using the :code:`context` for bucketing.

        If dict() is called on a Decision instance, the following fields are returned:
            .. code-block:: python
            {
                "variant": Optional[str],
                "value": Optional[JsonValue],
                "feature_id": int,
                "feature_name": str,
                "feature_version": int,
                "events": [str]
            }

        Raises :code:`DeciderException` if an error occurs.

        :param context: A required :code:`Mapping[str, JsonValue]` of context fields used for
            feature overrides, targeting, & bucketing.

        :param bucketing_field_filter: An optional string used to filter out features to be bucketed.
            Features with a "bucket_val" field that does not match the
            :code:`bucketing_field_filter` param will be excluded from bucketing.

        :return: A dict of Decision instances as the values and the corresponding
            feature name as the key.
        """
        if context is None:
            decider_client_counter.labels(
                operation="choose_all",
                success="false",
                error_type="missing_context",
                pkg_version=self._pkg_version,
            ).inc()
            raise DeciderException("Missing `context` param")

        ctx = make_ctx(context)
        ctx_err = ctx.err()
        if ctx_err:
            decider_client_counter.labels(
                operation="choose_all",
                success="false",
                error_type="invalid_context",
                pkg_version=self._pkg_version,
            ).inc()
            raise DeciderException(
                f"Encountered error from rust_decider.make_ctx() for context: {ctx_err}"
            )

        try:
            all_decisions = self._decider.choose_all(ctx, bucketing_field_filter)
        except Exception as e:
            decider_client_counter.labels(
                operation="choose_all",
                success="false",
                error_type="exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e

        output_decisions = {}

        for feature_name, decision in all_decisions.items():
            output_decisions[feature_name] = Decision(
                variant=decision.variant_name,
                feature_id=decision.feature_id,
                feature_name=decision.feature_name,
                feature_version=decision.feature_version,
                value=decision.value,
                events=decision.event_data,
                full_events=self._map_events(decision.events),
            )

        decider_client_counter.labels(
            operation="choose_all",
            success="true",
            error_type="",
            pkg_version=self._pkg_version,
        ).inc()

        return output_decisions

    def get_bool(self, feature_name: str, context: Mapping[str, JsonValue]) -> bool:
        """Fetch a Dynamic Configuration of boolean type.

        Raises :code:`DeciderException` if an error occurs when fetching DC.

        Raises :code:`ValueTypeMismatchException` (a sub-exception of :code:`DeciderException`)
        if requested feature is not of type boolean.

        :param feature_name: Name of the dynamic config you want a value for.

        :param context: A required :code:`Mapping[str, JsonValue]` of context fields used for
            feature overrides & targeting.

        :return: the boolean value of the dyanimc config if it is active/exists,
            raises a :code:`FeatureNotFoundException`
            (a sub-exception of :code:`DeciderException`) otherwise.
        """
        return self._get_dynamic_config_value(
            feature_name, context, self._decider.get_bool
        )

    def get_int(self, feature_name: str, context: Mapping[str, JsonValue]) -> int:
        """Fetch a Dynamic Configuration of int type.

        Raises :code:`DeciderException` if an error occurs when fetching DC.

        Raises :code:`ValueTypeMismatchException` (a sub-exception of :code:`DeciderException`)
        if requested feature is not of type integer.

        :param feature_name: Name of the dynamic config you want a value for.

        :param context: A required :code:`Mapping[str, JsonValue]` of context fields used for
            feature overrides & targeting.

        :return: the int value of the dyanimc config if it is active/exists,
            raises a :code:`FeatureNotFoundException`
            (a sub-exception of :code:`DeciderException`) otherwise.
        """
        return self._get_dynamic_config_value(
            feature_name, context, self._decider.get_int
        )

    def get_float(self, feature_name: str, context: Mapping[str, JsonValue]) -> float:
        """Fetch a Dynamic Configuration of float type.

        Raises :code:`DeciderException` if an error occurs when fetching DC.

        Raises :code:`ValueTypeMismatchException` (a sub-exception of :code:`DeciderException`)
        if requested feature is not of type float.

        :param feature_name: Name of the dynamic config you want a value for.

        :param context: A required :code:`Mapping[str, JsonValue]` of context fields used for
            feature overrides & targeting.

        :return: the float value of the dyanimc config if it is active/exists,
            raises a :code:`FeatureNotFoundException`
            (a sub-exception of :code:`DeciderException`) otherwise.
        """
        return self._get_dynamic_config_value(
            feature_name, context, self._decider.get_float
        )

    def get_string(self, feature_name: str, context: Mapping[str, JsonValue]) -> str:
        """Fetch a Dynamic Configuration of string type.

        Raises :code:`DeciderException` if an error occurs when fetching DC.

        Raises :code:`ValueTypeMismatchException` (a sub-exception of :code:`DeciderException`)
        if requested feature is not of type string.

        :param feature_name: Name of the dynamic config you want a value for.

        :param context: A required :code:`Mapping[str, JsonValue]` of context fields used for
            feature overrides & targeting.

        :return: the string value of the dyanimc config if it is active/exists,
            raises a :code:`FeatureNotFoundException`
            (a sub-exception of :code:`DeciderException`) otherwise.
        """
        return self._get_dynamic_config_value(
            feature_name, context, self._decider.get_string
        )

    def get_map(
        self, feature_name: str, context: Mapping[str, JsonValue]
    ) -> Optional[dict]:
        """Fetch a Dynamic Configuration of map type.

        Raises :code:`DeciderException` if an error occurs when fetching DC.

        Raises :code:`ValueTypeMismatchException` (a sub-exception of :code:`DeciderException`)
        if requested feature is not of type map.

        :param feature_name: Name of the dynamic config you want a value for.

        :param context: A required :code:`Mapping[str, JsonValue]` of context fields used for
            feature overrides & targeting.

        :return: the map value of the dyanimc config if it is active/exists,
            raises a :code:`FeatureNotFoundException`
            (a sub-exception of :code:`DeciderException`) otherwise.
        """
        return self._get_dynamic_config_value(
            feature_name, context, self._decider.get_map
        )

    def _get_dynamic_config_value(
        self,
        feature_name: str,
        context: Mapping[str, JsonValue],
        get_fn: Callable,
    ) -> Optional[Any]:
        dc_type = get_fn.__name__

        if context is None:
            decider_client_counter.labels(
                operation=f"{dc_type}",
                success="false",
                error_type="missing_context",
                pkg_version=self._pkg_version,
            ).inc()
            raise DeciderException(
                f"Missing `context` param for feature_name: {feature_name}"
            )

        ctx = make_ctx(context)
        ctx_err = ctx.err()
        if ctx_err:
            decider_client_counter.labels(
                operation=f"{dc_type}",
                success="false",
                error_type="invalid_context",
                pkg_version=self._pkg_version,
            ).inc()
            raise DeciderException(
                f"Encountered error from rust_decider.make_ctx(): {ctx_err}"
            )

        try:
            dc = get_fn(feature_name, ctx)
        except FeatureNotFoundException as e:
            decider_client_counter.labels(
                operation=f"{dc_type}",
                success="false",
                error_type="feature_not_found",
                pkg_version=self._pkg_version,
            ).inc()
            raise e
        except ValueTypeMismatchException as e:
            decider_client_counter.labels(
                operation=f"{dc_type}",
                success="false",
                error_type="type_mismatch",
                pkg_version=self._pkg_version,
            ).inc()
            raise e
        except DeciderException as e:
            decider_client_counter.labels(
                operation=f"{dc_type}",
                success="false",
                error_type="decider_exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e
        except Exception as e:
            decider_client_counter.labels(
                operation=f"{dc_type}",
                success="false",
                error_type="exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e

        decider_client_counter.labels(
            operation=f"{dc_type}",
            success="true",
            error_type="",
            pkg_version=self._pkg_version,
        ).inc()

        return dc

    def all_values(
        self,
        context: Mapping[str, JsonValue],
    ) -> Dict[str, JsonValue]:
        """Return a :code:`Dict[str, JsonValue]` for all active Dynamic Configurations...

        (with the key as the feature_name) using the :code:`context` for overrides/targeting.

        The value of the return dict contains the Dynamic Config value.

        Dynamic Configurations that are malformed, fail parsing, or otherwise
        error for any reason are included in the response and have their respective default
        values set:

        .. code-block:: python

            boolean -> False
            integer -> 0
            float   -> 0.0
            string  -> ""
            map     -> {}

        :param context: A required :code:`Mapping[str, JsonValue]` of context fields used for
            feature overrides & targeting.

        :return: A dict of Dynamic Config :code:`JsonValue`s as the dict values and
        the corresponding :code:`feature_name` as the key.
        """
        if context is None:
            decider_client_counter.labels(
                operation="all_values",
                success="false",
                error_type="missing_context",
                pkg_version=self._pkg_version,
            ).inc()
            raise DeciderException("Missing `context` param")

        ctx = make_ctx(context)
        ctx_err = ctx.err()
        if ctx_err:
            decider_client_counter.labels(
                operation="all_values",
                success="false",
                error_type="invalid_context",
                pkg_version=self._pkg_version,
            ).inc()
            raise DeciderException(
                f"Encountered error from rust_decider.make_ctx() for context: {ctx_err}"
            )

        try:
            all_decisions = self._decider.all_values(ctx)
        except Exception as e:
            decider_client_counter.labels(
                operation="all_values",
                success="false",
                error_type="exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e

        values = {}

        for feature_name, decision in all_decisions.items():
            values[feature_name] = decision.value

        decider_client_counter.labels(
            operation="all_values",
            success="true",
            error_type="",
            pkg_version=self._pkg_version,
        ).inc()

        return values

    def get_feature(self, feature_name: str) -> Optional[Feature]:
        """Get a :py:class:`~rust_decider.Feature` representation of a feature...

        if it's active/exists, otherwise raises :code:`FeatureNotFoundException`
        (a sub-exception of :code:`DeciderException`).

        :param feature_name: Name of the feature to be fetched.

        :return: a :py:class:`~reddit_decider.Feature` representation
            of a feature if found, raises :code:`FeatureNotFoundException`
            (a sub-exception of :code:`DeciderException`) otherwise.
        """
        try:
            feature = self._decider.get_feature(feature_name)
        except FeatureNotFoundException as e:
            decider_client_counter.labels(
                operation="get_feature",
                success="false",
                error_type="feature_not_found",
                pkg_version=self._pkg_version,
            ).inc()
            raise e
        except DeciderException as e:
            decider_client_counter.labels(
                operation="get_feature",
                success="false",
                error_type="decider_exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e
        except Exception as e:
            decider_client_counter.labels(
                operation="get_feature",
                success="false",
                error_type="exception",
                pkg_version=self._pkg_version,
            ).inc()
            raise e

        decider_client_counter.labels(
            operation="get_feature",
            success="true",
            error_type="",
            pkg_version=self._pkg_version,
        ).inc()

        return Feature(
            id=feature.id,
            name=feature.name,
            version=feature.version,
            bucket_val=feature.bucket_val,
            start_ts=feature.start_ts,
            stop_ts=feature.stop_ts,
            emit_event=feature.emit_event,
            measured=feature.measured,
        )



import os

TEST_DIR = os.path.dirname(os.path.abspath(__file__))

if __name__ == "__main__":
    import rust_decider

    decider = rust_decider.init(
        "darkmode overrides targeting holdout mutex_group fractional_availability value",
        f"{TEST_DIR}/../../cfg.json",
    )

    if decider:
        ctx_dict = {
            "user_id": "795244",
            "locale": None,
            "canonical_url": "www.reddit.com",
            "user_is_employee": True,
            "logged_in": False,
            "app_name": "ios",
            "build_number": 1234,
            "country_code": "UA",
            "cookie_created_timestamp": 1648859753.233,
            "other_fields": {
                "knoxness": "max",
                "romans": 7,
                "yiqi_is_badass": True,
                "romansFlaws": None,
            },
        }
        ctx = rust_decider.make_ctx(ctx_dict)
        print("context:")
        print(ctx.inspect())

        # choose
        a = decider.choose("genexp_0", ctx)
        print(a.inspect())
        print(f"'genexp_0' choose() error (should be None): {a.err()}")
        print(f"choose() decision(): {a.decision()}")
        print(f"choose() events(): {a.events()}")

        # choose bucket_val: "device_id", without device_id in ctx
        b = decider.choose("genexp_device_id", ctx)
        print(
            f"'genexp_device_id' choose() error (should be 'Missing device_id in context'): {b.err()}"
        )

        ca = decider.choose_all(ctx)
        print(f"results from choose_all: {ca}")

        # choose bucket_val: "device_id"
        ctx_dict = {"user_id": "blah", "device_id": "795244"}
        ctx = rust_decider.make_ctx(ctx_dict)
        c = decider.choose("genexp_device_id", ctx)
        print(f"'genexp_device_id' choose() error (shuold be None): {c.err()}")
        print(f"choose() decision(): {c.decision()}")

        # get_bool
        d = decider.get_bool("dc_bool", ctx)
        print(f"get_bool(): {d.val()}")

        # get_int
        e = decider.get_int("dc_int", ctx)
        print(f"get_int(): {e.val()}")

        # get_float
        f = decider.get_float("dc_float", ctx)
        print(f"get_float(): {f.val()}")

        # get_string
        g = decider.get_string("dc_string", ctx)
        print(f"get_string(): {g.val()}")

        # get_string missing feature error
        h = decider.get_string("missing", ctx)
        print(f"get_string() err: {h.err()}")

        # get_map
        i = decider.get_map("dc_map", ctx).val()
        print(f"get_map(): {i}")
        print(f"get_map() nested_map: {i['v']['nested_map']}")
        print(f"get_map() nested_map val: {i['v']['nested_map']['x']}")

        # get_experiment
        g = decider.get_experiment("genexp_0")
        print(f"get_experiment(): {g.err()}")
        print(f"get_experiment() {g.val()}")
    else:
        print("Failed to init decider")



#!/usr/bin/env python
import json

from unittest import TestCase
from unittest.mock import ANY
from unittest.mock import Mock
from unittest.mock import patch

from event_utils.v2_event_utils import ExperimentLogger
from event_utils.v2_event_utils import V2EventLogger
from rust_decider import Decider
from rust_decider.events import ExperimentLoggerExposerAdapter
from rust_decider.events import Exposer
from utils import APP_NAME
from utils import BASE_URL
from utils import BUILD_NUMBER
from utils import CANONICAL_URL
from utils import COOKIE_CREATED_TIMESTAMP
from utils import COUNTRY_CODE
from utils import create_temp_config_file
from utils import ctx_dict
from utils import DEVICE_ID
from utils import event_json
from utils import GEN_EXP_ID
from utils import GEN_EXP_NAME
from utils import GEN_EXP_VERSION
from utils import IS_EMPLOYEE
from utils import IS_LOGGED_IN
from utils import REFERRER_URL
from utils import SUBREDDIT_ID
from utils import USER_AGENT
from utils import USER_ID


class TestDeciderExpose(TestCase):
    def setUp(self):
        def __init__(self, event_queue=None):
            return

        with patch.object(V2EventLogger, "__init__", __init__):
            self.exp_logger = ExperimentLogger()
            self.exp_logger._v2_logger = Mock(spec=V2EventLogger)

        self.valid_ctx_dict = ctx_dict

        self.genexp_0_cfg = {
            "genexp_0": {
                "id": GEN_EXP_ID,
                "name": GEN_EXP_NAME,
                "enabled": True,
                "owner": "test",
                "version": GEN_EXP_VERSION,
                "emit_event": True,
                "type": "range_variant",
                "start_ts": 0,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [
                        {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
                        {"range_start": 0.2, "range_end": 0.4, "name": "variant_2"},
                        {"range_start": 0.4, "range_end": 0.6, "name": "variant_3"},
                        {"range_start": 0.6, "range_end": 0.8, "name": "variant_4"},
                        {"range_start": 0.8, "range_end": 1.0, "name": "variant_5"},
                    ],
                    "experiment_version": 5,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
        }

        super().setUp()

    def test_choose_with_exposer(self):
        mock_expose_fn = Mock()
        exposer = Exposer(expose_fn=mock_expose_fn)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = Decider(f.name, exposer)

            choice = decider.choose(
                feature_name="genexp_0", context=self.valid_ctx_dict, expose=True
            )

            assert len(choice.full_events) == 1
            mock_expose_fn.assert_called_once()
            logger_arg = mock_expose_fn.call_args.args[0]

            self.assertDictEqual(
                json.loads(logger_arg),
                event_json(
                    ctx=self.valid_ctx_dict,
                    noun="user_id",
                    variant="variant_5",
                    experiment_name=GEN_EXP_NAME,
                    experiment_id=GEN_EXP_ID,
                    is_override=False,
                ),
            )

    def test_choose_with_exposer_without_expose(self):
        mock_expose_fn = Mock()
        exposer = Exposer(expose_fn=mock_expose_fn)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = Decider(f.name, exposer)

            choice = decider.choose(
                feature_name="genexp_0", context=self.valid_ctx_dict, expose=False
            )

            assert len(choice.full_events) == 1
            mock_expose_fn.assert_not_called()

    def test_choose_with_exposer_for_holdout(self):
        mock_expose_fn = Mock()
        exposer = Exposer(expose_fn=mock_expose_fn)

        cfg = {
            "holdout": {
                "id": 1,
                "name": "holdout",
                "version": "5",
                "type": "range_variant",
                "enabled": True,
                "emit_event": True,
                "owner": "test",
                "experiment": {
                    "variants": [
                        {"name": "holdout", "range_end": 0.0, "range_start": 0.0},
                        {"name": "control_1", "range_end": 1.0, "range_start": 0.0},
                    ],
                    "experiment_version": 5,
                    "shuffle_version": 0,
                    "bucket_val": "user_id",
                },
            },
            "child": {
                "id": 2,
                "name": "child",
                "version": "1",
                "type": "range_variant",
                "enabled": True,
                "emit_event": True,
                "owner": "asdf",
                "parent_hg_name": "holdout",
                "experiment": {
                    "variants": [
                        {"name": "control_1", "range_end": 1.0, "range_start": 0.0}
                    ],
                    "experiment_version": 0,
                    "shuffle_version": 0,
                    "bucket_val": "user_id",
                },
            },
        }

        with create_temp_config_file(cfg) as f:
            decider = Decider(f.name, exposer)

            # only expose holdout events
            choice = decider.choose(
                feature_name="child",
                context=self.valid_ctx_dict,
                expose=False,
                expose_holdout=True,
            )

            # 2 events present
            assert len(choice.full_events) == 2
            # only expose holdout event
            mock_expose_fn.assert_called_once()
            logger_arg = mock_expose_fn.call_args.args[0]

            self.assertDictEqual(
                json.loads(logger_arg),
                event_json(
                    ctx=self.valid_ctx_dict,
                    noun="user_id",
                    variant="control_1",
                    experiment_name="holdout",
                    experiment_id=1,
                    is_override=False,
                ),
            )

    def test_choose_with_exposer_adapter(self):
        logger_adapter = ExperimentLoggerExposerAdapter(self.exp_logger)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = Decider(f.name, logger_adapter)

            choice = decider.choose(
                feature_name="genexp_0", context=self.valid_ctx_dict, expose=True
            )

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_5",
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [
                        f"0::::6299::::genexp_0::::5::::variant_5::::{USER_ID}::::user_id::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )

            assert len(choice.full_events) == 1
            logger_kwargs = logger_adapter._logger._v2_logger.log.call_args.kwargs

            self.assertDictEqual(
                logger_kwargs,
                {
                    "device_id": DEVICE_ID,
                    "user_id": USER_ID,
                    "build_number": BUILD_NUMBER,
                    "source": "experiment",
                    "action": "expose",
                    "noun": "user_id",
                    "uuid": ANY,
                    "client_timestamp": ANY,
                    "experiment": {
                        "id": GEN_EXP_ID,
                        "name": GEN_EXP_NAME,
                        "variant": "variant_5",
                        "owner": ANY,
                        "version": str(5),
                        "bucketing_key": "user_id",
                        "bucketing_value": USER_ID,
                        "end_timestamp": ANY,
                        "start_timestamp": ANY,
                    },
                    "user": {
                        "id": USER_ID,
                        "logged_in": IS_LOGGED_IN,
                        "cookie_created_timestamp": COOKIE_CREATED_TIMESTAMP,
                        "is_employee": IS_EMPLOYEE,
                        "created_timestamp": ANY,
                    },
                    "app": {
                        "name": APP_NAME,
                        "build_number": BUILD_NUMBER,
                    },
                    "platform": {"device_id": DEVICE_ID},
                    "geo": {"country_code": COUNTRY_CODE},
                    "request": {
                        "canonical_url": CANONICAL_URL,
                        "base_url": BASE_URL,
                        "user_agent": USER_AGENT,
                    },
                    "referrer": {"url": REFERRER_URL},
                    "base_url": BASE_URL,
                    "user_agent": USER_AGENT,
                    "subreddit": {"id": SUBREDDIT_ID},
                    "app_name": APP_NAME,
                    "canonical_url": CANONICAL_URL,
                    "subreddit_id": SUBREDDIT_ID,
                    "cookie_created_timestamp": COOKIE_CREATED_TIMESTAMP,
                    "correlation_id": ANY,
                    "country_code": COUNTRY_CODE,
                    "logged_in": IS_LOGGED_IN,
                    "user_is_employee": IS_EMPLOYEE,
                    "zipkin": ANY,
                },
            )

    def test_choose_with_exposer_adapter_without_expose(self):
        logger_adapter = ExperimentLoggerExposerAdapter(self.exp_logger)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = Decider(f.name, logger_adapter)

            choice = decider.choose(
                feature_name="genexp_0", context=self.valid_ctx_dict, expose=False
            )

            self.assertEqual(
                dict(choice),
                {
                    "variant": "variant_5",
                    "value": None,
                    "feature_id": 6299,
                    "feature_name": "genexp_0",
                    "feature_version": 5,
                    "events": [
                        f"0::::6299::::genexp_0::::5::::variant_5::::{USER_ID}::::user_id::::0::::2147483648"
                    ],
                    "full_events": ANY,
                },
            )

            assert len(choice.full_events) == 1
            logger_adapter._logger._v2_logger.log.assert_not_called()



#!/usr/bin/env python
import json
import os
import random
import time

from unittest import TestCase

import rust_decider

from hypothesis import strategies as st
from reddit_experiments import Experiments
from utils import APP_NAMES
from utils import LOCALES
from utils import make_experiment
from utils import make_request_context_map


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def setup_baseplate_experiments(cfg_path):
    h = json.load(open(cfg_path))
    # we really only want to populate the raw string cache of _cfg_data,
    # so that when we go looking for keys, we find them.
    el = Experiments(
        cfg_data=h, config_watcher=None, server_span=None, context_name=None
    )
    return el


def setup_decider(cfg_path):
    return rust_decider.init(
        "darkmode overrides targeting holdout mutex_group fractional_availability",
        cfg_path,
    )


@st.composite
def make_ctx_map(draw):
    return {  # TODO: make the nullable fields sometimes be None
        "user_id": str(draw(st.integers(min_value=1, max_value=1000))),
        "locale": draw(st.sampled_from(LOCALES + [None])),
        "device_id": str(draw(st.integers(min_value=10000, max_value=11000))),
        "country_code": draw(st.sampled_from(LOCALES + [None])),
        "origin_service": "shreddit",  # do I care about this one?
        "user_is_employee": draw(st.booleans()),
        "logged_in": draw(st.booleans()),
        "app_name": draw(st.sampled_from(APP_NAMES + [None])),
        "build_number": str(draw(st.integers(min_value=1, max_value=1000))),
    }


class Exp:
    def __init__(
        self,
        eid,
        name,
        enabled,
        version,
        shuffle_version,
        bucket_val,
        targeting=None,
        overrides=None,
    ):
        self.id = eid
        self.name = name
        self.enabled = enabled
        self.version = str(version)
        self.type = "range_variant"
        self.start_ts = 0
        self.stop_ts = 2147483648
        self.experiment = {
            "variants": make_variants(),
            "experiment_version": int(version),
            "shuffle_version": int(shuffle_version),
            "bucket_val": bucket_val,
            "log_bucketing": False,
            "overrides": overrides,
            "targeting": targeting,
        }

    @classmethod
    def strategy(cls):
        return st.builds(
            cls,
            st.integers(min_value=1),
            st.from_regex(r"^\A[a-z_0-9]{5,40}\Z"),
            st.booleans(),
            st.integers(min_value=0, max_value=10000),
            st.integers(min_value=0),
            st.sampled_from(["user_id", "device_id"]),
            targeting_tree(),
        )

    @classmethod
    def from_strategy(cls, s):
        return cls(
            s.id,
            s.name,
            s.enabled,
            s.version,
            s.shuffle_version,
            s.bucket_val,
            s.targeting,
            s.overrides,
        )

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "enabled": self.enabled,
            "version": self.version,
            "owner": "test",
            "type": "range_variant",
            "start_ts": 0,  # we'Re not interested in testing start/stop
            "stop_ts": 2147483648,
            "experiment": {
                "variants": self.experiment["variants"],
                "experiment_version": self.experiment["experiment_version"],
                "shuffle_version": self.experiment["shuffle_version"],
                "bucket_val": self.experiment["bucket_val"],
                "log_bucketing": False,
                # "overrides": {}, # TODO: build this.
                "targeting": self.experiment["targeting"],
            },
        }


@st.composite
def make_exp(draw):
    version = draw(st.integers(min_value=0, max_value=10000))
    variants = make_variants()
    return {
        "id": draw(st.integers(min_value=1)),
        "name": draw(st.from_regex(r"^\A[a-z_0-9]{5,40}\Z")),
        "enabled": True,  # TODO: make this false sometimes?  lowpri
        "owner": "test",
        "version": str(version),
        "type": "range_variant",
        "start_ts": 0,  # we'Re not interested in testing start/stop
        "stop_ts": 2147483648,
        "experiment": {
            "variants": variants,
            "experiment_version": int(version),
            "shuffle_version": draw(st.integers(min_value=0)),
            "bucket_val": "user_id",
            "log_bucketing": False,
            # "overrides": {}, # TODO: build this.
            "targeting": targeting_tree(),
        },
    }


def make_op_node(op, field, val):
    """Generate a targeting tree op node."""
    if isinstance(val, list):
        val_field = "values"
    else:
        val_field = "value"

    val = str(val) if isinstance(val, bool) else val

    return {op: {"field": field, val_field: val}}


@st.composite
def targeting_tree(draw):
    return draw(
        st.recursive(
            bool_eq() | str_eq(),
            lambda cl: not_n(cl) | any_n(cl) | all_n(cl),
            max_leaves=20,
        )
    )


@st.composite
def user_ids(draw, elements=st.integers(min_value=1, max_value=1000)):
    return draw(st.lists(elements, min_size=1))


@st.composite
def bool_eq(draw):
    field = draw(st.sampled_from(["user_is_employee", "logged_in"]))
    val = draw(st.booleans())
    return make_op_node("EQ", field, val)


@st.composite
def str_eq(draw):
    field = draw(st.sampled_from(["user_id", "device_id"]))
    val = draw(user_ids())
    return make_op_node("EQ", field, val)


@st.composite
def not_n(draw, children):
    cl = draw(children)
    return {"NOT": cl}


@st.composite
def all_n(draw, children):
    cl = draw(st.lists(children))
    return {"ALL": cl}  # maybe have to set min_size?


@st.composite
def any_n(draw, children):
    cl = draw(st.lists(children))
    return {"ANY": cl}  # maybe have to set min_size?


def make_variants(h={}):
    return h or [  # TODO: actually generate variantsets.  Lowpri?
        {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
        {"range_start": 0.2, "range_end": 0.4, "name": "variant_2"},
        {"range_start": 0.4, "range_end": 0.6, "name": "variant_3"},
        {"range_start": 0.6, "range_end": 0.8, "name": "variant_4"},
    ]


def compare_implementations_against_cfg(
    cfg_path, user_ids, exp_names=None, ctx=make_request_context_map()
):
    if exp_names is None:
        h = json.load(open(cfg_path))

        # filter out DCs from the  default cfg- BP can't handle them
        exp_names = [k for k in h.keys() if k[0] in ["e", "g"]]

    dc = setup_decider(cfg_path)
    bp = setup_baseplate_experiments(cfg_path)
    successes = []
    failures = []
    for exp_name in exp_names:
        for uid in user_ids:
            ctx["user_id"] = str(uid)
            res, bpv = compare_single_call(dc, bp, exp_name, uid, ctx)
            if res.err() or not (bpv == res.decision()):
                failures.append(
                    {
                        "uid": uid,
                        "exp_name": exp_name,
                        "dc": res.decision(),
                        "bpv": bpv,
                        "ctx": ctx,
                        "res": res,
                    }
                )
                print(
                    f"failure:{exp_name} {uid} bp:{bpv} != dc:{res.decision()} ctx={ctx}"
                )
            else:
                successes.append(uid)
    return successes, failures


def compare_single_call(dc, bp, exp_name, uid, ctx_map=None):
    ctx_map = ctx_map or make_request_context_map({"user_id": str(uid)})
    ctx = rust_decider.make_ctx(ctx_map)
    res = dc.choose(exp_name, ctx)
    bpv = bp.variant(exp_name, **ctx_map)
    return [res, bpv]


class TestRustWithSdk(TestCase):
    def test_impls_against_cfg(self):
        cfg_path = f"{TEST_DIR}/../../cfg.json"
        h = json.load(open(cfg_path))
        exp_names = [k for k in h.keys() if k[0] in ["e", "g"]]
        uids = range(3000)
        good, bad = compare_implementations_against_cfg(cfg_path, uids, exp_names)
        if bad:
            print("mismatches found!\n\n", bad)
        else:
            print("all_good!")

        self.assertEqual(bad, [])
        self.assertEqual(len(good), len(uids) * len(exp_names))

    def test_impls_against_x0(self):
        cfg_path = f"{TEST_DIR}/../../cfg.json"
        exp_names = ["x0"]
        uids = range(100)
        ctx = make_request_context_map({"user_id": str(1), "user_is_employee": True})
        good, bad = compare_implementations_against_cfg(cfg_path, uids, exp_names, ctx)
        self.assertEqual(bad, [])
        self.assertEqual(len(good), len(uids) * len(exp_names))

    def test_impls_against_generated(self):
        # we want to be able to reproduce failures, so seed the RNG with the current timestamp,
        # and then log it on failures.  This will assist in debugging test fails.
        seed = int(time.time())
        random.seed(seed)
        msg = f"random seed={seed}"

        exp_cfg = {e["name"]: e for e in [make_experiment(n) for n in range(20)]}
        fn = "/tmp/genexp"  # TODO: generate the filename
        f = open(fn, "w")
        f.write(json.dumps(exp_cfg, indent=2))
        f.close()
        good, bad = compare_implementations_against_cfg(fn, range(3000))
        self.assertEqual(len(bad), 0, msg)
        self.assertEqual(len(good), 3000 * 20, msg)

    # @given(Exp.strategy())
    # def test_impls_against_hypothesis(self, exp):
    #     seed = int(time.time())
    #     random.seed(seed)
    #     msg = f"random seed={seed}"
    #     print(exp.to_dict())
    #     exp_cfg = {e["name"]:e for e in [exp.to_dict()]}
    #     fn = "/tmp/hypexp" # TODO: generate the filename
    #     f = open(fn, "w")
    #     f.write(json.dumps(exp_cfg, indent=2))
    #     f.close()
    #     good, bad = compare_implementations_against_cfg(fn, range(3000))
    #     self.assertEqual(len(bad), 0, msg)
    #     self.assertEqual(len(good), 3000, msg)



from setuptools import find_packages
from setuptools import setup

setup(
    name="reddit-decider",
    description="Reddit's python experiments framework",
    long_description="""
    Reddit's Python experiments framework.
    Bucketing, targeting, overrides, and dynamic config logic is implemented in Rust and wrapped in this Python package.
    """,
    long_description_content_type="text/markdown",
    url="https://github.snooguts.net/reddit/decider-py",
    project_urls={
        "Documentation": "https://reddit-experiments.readthedocs.io/",
    },
    author="matt knox",
    author_email="matt.knox@reddit.com",
    license="MIT",
    use_scm_version=True,
    packages=find_packages(),
    python_requires=">=3.7",
    setup_requires=["setuptools_scm"],
    install_requires=["reddit-edgecontext>=1.0.0a3,<2.0", "maturin>=0.10,<0.13"],
    package_data={"reddit_experiments": ["py.typed"]},
    zip_safe=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
    ],
)



import json

from abc import ABC
from abc import abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from typing import Callable
from typing import Optional


class EventType(Enum):
    EXPOSE = "expose"


@dataclass
class Experiment:
    id: int
    version: str
    name: str
    variant: str
    bucket_val: str  # corresponds to bucketing_key in v2 event schema
    start_ts: Optional[int]
    stop_ts: Optional[int]
    bucketing_value: str
    is_override: Optional[bool]


class ExposerABC(ABC):
    def __init__(self, expose_fn: Callable):
        self._expose_fn = expose_fn

    @abstractmethod
    def expose_event(self, event: str) -> None:
        pass


class Exposer(ExposerABC):
    def expose_event(self, event: str) -> None:
        if event and self._expose_fn:
            self._expose_fn(event)


class ExperimentLoggerExposerAdapter(Exposer):
    def __init__(self, logger):
        self._logger = logger

    def expose_event(self, event: str):
        if event and self._logger:
            event_dict = self._make_event_dict(event)
            self._logger.log(**event_dict)

    def _make_event_dict(self, event_str: str) -> dict:
        """
        Convert thrift event schema json string into dict used as params by ExperimentLogger.

        e.g. turns:
            {
                "1": {"str": "experiment"},
                "2": {"str": "expose"},
                "3": {"str": "user_id"},
                "107": {
                    "rec": {
                        "2": {"str": "ios"},
                        "4": {"i32": 1234},
                        "6": {"str": "us_en"
                    }
                }
                ...
            }
        into:
            {
                "source": "experiment",
                "action": "expose",
                "noun": "user_id",
                "app": {
                    "name": "ios",
                    "build_number": 1234,
                    "relevant_locale": "us_en"
                }
                ...
            }

        Based around the ExperimentLogger definition that can be found here:
            https://github.snooguts.net/reddit/data-schemas/blob/master/thrift_packages/py/event_utils/v2_event_utils.py
        ExperimentLogger has been copy/pasted across many repos with some modifications for which fields get exposed.

        Those modifications are no longer supported, since the expose event has been pruned to this protobuf schema:
            https://github.snooguts.net/reddit/event-schema-protos/blob/main/proto/events/experiments-team/experiment__expose.proto
        and only fields that are defined in Decider's context and propogated from the Decision
        have been mapped to their thrift json serialized v2 event schema counterpart.
        """
        event_fields = {}
        event = json.loads(event_str)
        for e_num, e_schema in event.items():
            e_map = EVENT_FIELD_MAPPING.get(e_num)
            if e_map:
                event_fields[e_map["field_name"]] = e_map["schema_to_struct_fn"](
                    e_schema
                )

        experiment = event_fields.pop("experiment")
        full_fields = self._add_flattened_fields(event_fields)
        full_fields = {
            **full_fields,
            **{experiment.bucket_val: experiment.bucketing_value},
        }

        return dict(
            experiment=experiment,
            variant=experiment.variant,
            span=None,
            event_type=EventType.EXPOSE,
            inputs=full_fields,
            **full_fields,
        )

    def _add_flattened_fields(self, event_fields: dict) -> dict:
        """
        Add fields that are accessed as top-level kwargs, or from "inputs" kwarg.

        See here:
            https://github.snooguts.net/reddit/data-schemas/blob/36d11b5c9519b48b2c49b5adda23cbe9a95b8875/thrift_packages/py/event_utils/v2_event_utils.py#L137-L138
        """
        inputs = deepcopy(event_fields)

        app_struct = inputs.get("app", {})

        app_name = app_struct.get("name")
        if app_name:
            inputs["app_name"] = app_name

        build_number = app_struct.get("build_number")
        if build_number:
            inputs["build_number"] = build_number

        user_is_employee = inputs.get("user", {}).get("is_employee")
        if user_is_employee:
            inputs["user_is_employee"] = user_is_employee

        subreddit_id = inputs.get("subreddit", {}).get("id")
        if subreddit_id:
            inputs["subreddit_id"] = subreddit_id

        user_dict = inputs.get("user", {})
        user_dict.pop("id", None)
        user_dict.pop("is_employee", None)

        return {
            **inputs,
            **inputs.get("platform", {}),
            **inputs.get("request", {}),
            **inputs.get("geo", {}),
            **user_dict,
        }


def parse_experiment_schema(experiment_schema: dict) -> Experiment:
    exp_data = experiment_schema["rec"]
    is_override = exp_data.get("10", {}).get("tf")

    return Experiment(
        id=exp_data["1"]["i64"],
        name=exp_data["2"]["str"],
        variant=exp_data["4"]["str"],
        start_ts=exp_data.get("5", {}).get("i64"),
        stop_ts=exp_data.get("6", {}).get("i64"),
        bucket_val=exp_data["7"]["str"],
        version=exp_data["8"]["str"],
        bucketing_value=exp_data["9"]["str"],
        is_override=bool(is_override) if is_override is not None else None,
    )


def parse_top_level_str_schema(schema: dict):
    return schema["str"]


def parse_top_level_i64_schema(schema: dict):
    return schema["i64"]


def event_mapping_dict(field_name: str, parsing_fn: Callable):
    return {"field_name": field_name, "schema_to_struct_fn": parsing_fn}


def parse_app_schema(app_schema: dict) -> dict:
    app_data = app_schema["rec"]
    app_struct = {}

    name = app_data.get("2", {}).get("str")
    if name:
        app_struct["name"] = name

    build_number = app_data.get("4", {}).get("i32")
    if build_number:
        app_struct["build_number"] = build_number

    relevant_locale = app_data.get("6", {}).get("str")
    if relevant_locale:
        app_struct["relevant_locale"] = relevant_locale

    return app_struct


def parse_platform_schema(platform_schema: dict) -> dict:
    platform_data = platform_schema["rec"]
    platform_struct = {}

    device_id = platform_data.get("2", {}).get("str")
    if device_id:
        platform_struct["device_id"] = device_id

    return platform_struct


def parse_request_schema(request_schema: dict) -> dict:
    request_data = request_schema["rec"]
    request_struct = {}

    canonical_url = request_data.get("17", {}).get("str")
    if canonical_url:
        request_struct["canonical_url"] = canonical_url

    base_url = request_data.get("3", {}).get("str")
    if base_url:
        request_struct["base_url"] = base_url

    user_agent = request_data.get("1", {}).get("str")
    if user_agent:
        request_struct["user_agent"] = user_agent

    return request_struct


def parse_referrer_schema(referrer_schema: dict) -> dict:
    referrer_data = referrer_schema["rec"]
    referrer_struct = {}

    url = referrer_data.get("2", {}).get("str")
    if url:
        referrer_struct["url"] = url

    return referrer_struct


def parse_user_schema(user_schema: dict) -> dict:
    user_data = user_schema["rec"]
    user_struct = {}

    user_id = user_data.get("1", {}).get("str")
    if user_id:
        user_struct["user_id"] = user_id

    logged_in = user_data.get("3", {}).get("tf")
    if logged_in is not None:
        user_struct["logged_in"] = bool(logged_in)

    cookie_created_timestamp = user_data.get("4", {}).get("i64")
    if cookie_created_timestamp:
        user_struct["cookie_created_timestamp"] = cookie_created_timestamp

    is_employee = user_data.get("16", {}).get("tf")
    if is_employee is not None:
        user_struct["is_employee"] = bool(is_employee)

    return user_struct


def parse_subreddit_schema(subreddit_schema: dict) -> dict:
    subreddit_data = subreddit_schema["rec"]
    subreddit_struct = {}

    id = subreddit_data.get("1", {}).get("str")
    if id:
        subreddit_struct["id"] = id

    return subreddit_struct


def parse_geo_schema(geo_schema: dict) -> dict:
    geo_data = geo_schema["rec"]
    geo_struct = {}

    country_code = geo_data.get("1", {}).get("str")
    if country_code:
        geo_struct["country_code"] = country_code

    return geo_struct


EVENT_FIELD_MAPPING = {
    "1": event_mapping_dict("source", parse_top_level_str_schema),
    "2": event_mapping_dict("action", parse_top_level_str_schema),
    "3": event_mapping_dict("noun", parse_top_level_str_schema),
    "5": event_mapping_dict("client_timestamp", parse_top_level_i64_schema),
    "6": event_mapping_dict("uuid", parse_top_level_str_schema),
    "8": event_mapping_dict("correlation_id", parse_top_level_str_schema),
    "107": event_mapping_dict("app", parse_app_schema),
    "108": event_mapping_dict("platform", parse_platform_schema),
    "109": event_mapping_dict("request", parse_request_schema),
    "110": event_mapping_dict("referrer", parse_referrer_schema),
    "112": event_mapping_dict("user", parse_user_schema),
    "114": event_mapping_dict("subreddit", parse_subreddit_schema),
    "129": event_mapping_dict("experiment", parse_experiment_schema),
    "500": event_mapping_dict("geo", parse_geo_schema),
}



from prometheus_client import Counter

decider_client_counter = Counter(
    "decider_client_total",
    "Count of successful/failed Decider operations (with error_type) in reddit-decider package",
    ["operation", "success", "error_type", "pkg_version"],
)



#!/usr/bin/env python
import os

from contextlib import contextmanager
from unittest import TestCase

import rust_decider

from utils import create_temp_config_file

TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def setup_decider(cfg_path):
    return rust_decider.init(
        "darkmode overrides targeting holdout mutex_group fractional_availability value",
        cfg_path,
    )


class TestDeciderPy(TestCase):
    valid_ctx_dict = {
        "user_id": "795244",
        "device_id": "1234",
        "canonical_url": "www.reddit.com",
        "locale": "us_en",
        "user_is_employee": True,
        "logged_in": None,
        "app_name": "ios",
        "build_number": 1234,
        "country_code": "UA",
        "origin_service": "oss",
        "oauth_client_id": "test",
        "cookie_created_timestamp": 1648859753,
    }

    variants = [
        {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
        {"range_start": 0.2, "range_end": 0.4, "name": "variant_2"},
        {"range_start": 0.4, "range_end": 0.6, "name": "variant_3"},
        {"range_start": 0.6, "range_end": 0.8, "name": "variant_4"},
        {"range_start": 0.8, "range_end": 1.0, "name": "variant_5"},
    ]

    device_id_exp = {
        "genexp_device_id": {
            "id": 6222,
            "name": "genexp_device_id",
            "enabled": True,
            "owner": "test",
            "version": "5",
            "type": "range_variant",
            "start_ts": 0,
            "stop_ts": 2147483648,
            "emit_event": True,
            "experiment": {
                "variants": variants,
                "experiment_version": 5,
                "shuffle_version": 91,
                "bucket_val": "device_id",
                "log_bucketing": False,
            },
        }
    }

    canonical_url_exp = {
        "genexp_canonical_url": {
            "id": 6233,
            "name": "genexp_canonical_url",
            "enabled": True,
            "owner": "test",
            "version": "5",
            "type": "range_variant",
            "start_ts": 0,
            "stop_ts": 2147483648,
            "emit_event": True,
            "experiment": {
                "variants": variants,
                "experiment_version": 5,
                "shuffle_version": 91,
                "bucket_val": "canonical_url",
                "log_bucketing": False,
            },
        }
    }

    def setUp(self):
        super().setUp()
        self.ctx = rust_decider.make_ctx(self.valid_ctx_dict)
        self.genexp_0_cfg = {
            "genexp_0": {
                "id": 6299,
                "name": "genexp_0",
                "enabled": True,
                "owner": "test",
                "version": "5",
                "emit_event": True,
                "type": "range_variant",
                "start_ts": 0,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": self.variants,
                    "experiment_version": 5,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
        }
        self.additional_2_exp = {
            "exp_0": {
                "id": 3248,
                "name": "exp_0",
                "enabled": True,
                "owner": "test",
                "version": "2",
                "type": "range_variant",
                "emit_event": True,
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [
                        {"range_start": 0.0, "range_end": 0.2, "name": "control_1"},
                        {"range_start": 0.2, "range_end": 0.4, "name": "control_2"},
                        {"range_start": 0.4, "range_end": 0.6, "name": "variant_2"},
                        {"range_start": 0.6, "range_end": 0.8, "name": "variant_3"},
                        {"range_start": 0.8, "range_end": 1.0, "name": "variant_4"},
                    ],
                    "experiment_version": 2,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
            "exp_1": {
                "id": 3246,
                "name": "exp_1",
                "enabled": True,
                "owner": "test",
                "version": "2",
                "type": "range_variant",
                "emit_event": True,
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [
                        {"range_start": 0, "range_end": 0, "name": "variant_0"}
                    ],
                    "experiment_version": 2,
                    "shuffle_version": 0,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            },
        }

    def test_init(self):
        # handles full cfg.json file
        decider = setup_decider(f"{TEST_DIR}/../../cfg.json")
        self.assertEqual(decider.err(), None)

    def test_init_bad_cfg(self):
        # an experiment's id is string instead of int
        cfg = {
            "exp_0": {
                "id": "3248",
                "name": "exp_0",
                "enabled": True,
                "owner": "test",
                "version": "2",
                "type": "range_variant",
                "start_ts": 37173982,
                "stop_ts": 2147483648,
                "experiment": {
                    "variants": [],
                    "experiment_version": 2,
                    "shuffle_version": 91,
                    "bucket_val": "user_id",
                    "log_bucketing": False,
                },
            }
        }

        with create_temp_config_file(cfg) as f:
            decider = setup_decider(f.name)

            assert (
                'Decider initialization failed: Partially loaded Decider: 1 feature(s) failed to load: {"exp_0": ParsingError(Error("invalid type: string \\"3248\\", expected u32"'
                in decider.err()
            )

    def test_make_ctx(self):
        ctx = rust_decider.make_ctx(self.valid_ctx_dict)
        self.assertEqual(ctx.err(), None)

    def test_make_ctx_string_type_mismatch(self):
        str_fields = [
            "locale",
            "country_code",
            "app_name",
            "device_id",
            "canonical_url",
            "origin_service",
            "oauth_client_id",
        ]
        for str_field in str_fields:
            v_ctx = self.valid_ctx_dict.copy()
            v_ctx[str_field] = 1
            ctx = rust_decider.make_ctx(v_ctx)
            self.assertEqual(f'"{str_field}" type mismatch (string).', ctx.err())

    def test_make_ctx_bool_type_mismatch(self):
        bool_fields = ["logged_in", "user_is_employee"]
        for bool_field in bool_fields:
            v_ctx = self.valid_ctx_dict.copy()
            v_ctx[bool_field] = "not bool"
            ctx = rust_decider.make_ctx(v_ctx)
            self.assertEqual(f'"{bool_field}" type mismatch (bool).', ctx.err())

    def test_make_ctx_int_type_mismatch(self):
        int_fields = [
            "build_number",
            "cookie_created_timestamp",
        ]
        for int_field in int_fields:
            v_ctx = self.valid_ctx_dict.copy()
            v_ctx[int_field] = "not int"
            ctx = rust_decider.make_ctx(v_ctx)
            self.assertEqual(f'"{int_field}" type mismatch (integer).', ctx.err())

    # todo:
    # def test_make_ctx_without_user_id(self):

    # todo:
    # def test_make_ctx_with_None_fields(self):

    def test_choose(self):
        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider(f.name)

            choice = decider.choose("genexp_0", self.ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.decision(), "variant_5")
            self.assertEqual(
                choice.events(),
                [
                    "0::::6299::::genexp_0::::5::::variant_5::::795244::::user_id::::0::::2147483648"
                ],
            )

    def test_choose_bad_init(self):
        decider = rust_decider.init("foo", "bah")
        choice = decider.choose("bar", self.ctx)

        self.assertEqual(choice.err(), "Decider not found.")

    def test_choose_bucket_val_device_id(self):
        with create_temp_config_file(self.device_id_exp) as f:
            decider = setup_decider(f.name)

            choice = decider.choose("genexp_device_id", self.ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.decision(), "variant_5")

    def test_choose_bucket_val_device_id_missing_identifier(self):
        with create_temp_config_file(self.device_id_exp) as f:
            decider = setup_decider(f.name)
            ctx = self.valid_ctx_dict.copy()
            del ctx["device_id"]
            missing_device_id_ctx = rust_decider.make_ctx(ctx)

            choice = decider.choose("genexp_device_id", missing_device_id_ctx)

            self.assertEqual(
                choice.err(),
                'Missing field "device_id" in context for bucket_val = device_id',
            )
            self.assertEqual(choice.decision(), None)

    def test_choose_bucket_val_canonical_url(self):
        with create_temp_config_file(self.canonical_url_exp) as f:
            decider = setup_decider(f.name)

            choice = decider.choose("genexp_canonical_url", self.ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.decision(), "control_1")

    def test_choose_bucket_val_canonical_url_missing_identifier(self):
        with create_temp_config_file(self.canonical_url_exp) as f:
            decider = setup_decider(f.name)
            ctx = self.valid_ctx_dict.copy()
            del ctx["canonical_url"]
            missing_canonical_url_ctx = rust_decider.make_ctx(ctx)

            choice = decider.choose("genexp_canonical_url", missing_canonical_url_ctx)

            self.assertEqual(
                choice.err(),
                'Missing field "canonical_url" in context for bucket_val = canonical_url',
            )
            self.assertEqual(choice.decision(), None)

    def test_choose_bucket_val_canonical_url_with_identifer_type(self):
        with create_temp_config_file(self.canonical_url_exp) as f:
            decider = setup_decider(f.name)

            choice = decider.choose(
                feature_name="genexp_canonical_url",
                ctx=self.ctx,
                identifier_type="canonical_url",
            )

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.decision(), "control_1")

    def test_choose_bucket_val_canonical_url_with_mismatched_identifer_type(self):
        with create_temp_config_file(self.canonical_url_exp) as f:
            decider = setup_decider(f.name)

            choice = decider.choose(
                feature_name="genexp_canonical_url",
                ctx=self.ctx,
                identifier_type="device_id",
            )

            self.assertEqual(
                choice.err(),
                'Requested identifier_type "device_id" is incompatible with experiment\'s bucket_val = canonical_url',
            )
            self.assertEqual(choice.decision(), None)

    def test_choose_bucket_val_canonical_url_with_incorrect_identifer_type(self):
        with create_temp_config_file(self.canonical_url_exp) as f:
            decider = setup_decider(f.name)

            choice = decider.choose(
                feature_name="genexp_canonical_url",
                ctx=self.ctx,
                identifier_type="blah",
            )

            self.assertEqual(
                choice.err(),
                'Requested identifier_type "blah" is incompatible with experiment\'s bucket_val = canonical_url',
            )
            self.assertEqual(choice.decision(), None)

    def test_choose_with_other_fields_for_targeting(self):
        cfg = self.genexp_0_cfg.copy()
        cfg["genexp_0"]["experiment"].update(
            {"targeting": {"ALL": [{"EQ": {"field": "foo", "values": ["bar"]}}]}}
        )

        with create_temp_config_file(cfg) as f:
            decider = setup_decider(f.name)
            py_dict = self.valid_ctx_dict.copy()

            # targeting matches
            py_dict.update({"other_fields": {"foo": "bar"}})
            ctx = rust_decider.make_ctx(py_dict)

            choice = decider.choose("genexp_0", ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.decision(), "variant_5")
            self.assertEqual(
                choice.events(),
                [
                    "0::::6299::::genexp_0::::5::::variant_5::::795244::::user_id::::0::::2147483648"
                ],
            )

            # targeting doesn't match
            py_dict.update({"other_fields": {"foo": "huh"}})
            ctx = rust_decider.make_ctx(py_dict)

            choice = decider.choose("genexp_0", ctx)
            self.assertEqual(choice.decision(), None)

    @contextmanager
    def test_feature_not_found(self):
        with create_temp_config_file({}) as f:
            decider = setup_decider(f.name)
            yield decider

    def test_choose_feature_not_found(self):
        with self.test_feature_not_found() as d:
            result = d.choose("any", self.ctx)

        self.assertEqual(result.err(), 'Feature "any" not found')

    def test_choose_all(self):
        self.genexp_0_cfg.update(self.additional_2_exp)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider(f.name)

            choice_res = decider.choose_all(self.ctx)

            assert choice_res.err() is None
            choice_dict = choice_res.decisions()

            # assert genexp_0
            self.assertEqual(len(choice_dict), len(self.genexp_0_cfg))
            self.assertEqual(
                choice_dict["genexp_0"].decision_dict(),
                {
                    "name": "variant_5",
                    "version": "5",
                    "id": "6299",
                    "experimentName": "genexp_0",
                },
            )
            self.assertEqual(
                choice_dict["genexp_0"].events(),
                [
                    "0::::6299::::genexp_0::::5::::variant_5::::795244::::user_id::::0::::2147483648"
                ],
            )

            # assert exp_0
            self.assertEqual(
                choice_dict["exp_0"].decision_dict(),
                {
                    "name": "variant_3",
                    "version": "2",
                    "id": "3248",
                    "experimentName": "exp_0",
                },
            )
            self.assertEqual(
                choice_dict["exp_0"].events(),
                [
                    "0::::3248::::exp_0::::2::::variant_3::::795244::::user_id::::37173982::::2147483648"
                ],
            )

            # assert exp_1
            self.assertEqual(choice_dict["exp_1"].decision_dict(), {})
            self.assertEqual(choice_dict["exp_1"].events(), [])

    def test_choose_all_with_single_identifier_type(self):
        self.genexp_0_cfg.update(self.additional_2_exp)
        self.genexp_0_cfg.update(self.canonical_url_exp)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider(f.name)

            choice_res = decider.choose_all(self.ctx, identifier_type="canonical_url")

            assert choice_res.err() is None
            choice_dict = choice_res.decisions()

            # assert canonical_url_exp
            self.assertEqual(len(choice_dict), 1)
            self.assertEqual(
                choice_dict["genexp_canonical_url"].decision_dict(),
                {
                    "name": "control_1",
                    "version": "5",
                    "id": "6233",
                    "experimentName": "genexp_canonical_url",
                },
            )
            self.assertEqual(
                choice_dict["genexp_canonical_url"].events(),
                [
                    "0::::6233::::genexp_canonical_url::::5::::control_1::::www.reddit.com::::canonical_url::::0::::2147483648"
                ],
            )

    def test_get_bool(self):
        bool_val = True
        cfg = {
            "dc_bool": {
                "id": 3393,
                "value": bool_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_bool",
                "value_type": "Boolean",
                "experiment": {"experiment_version": 2},
            }
        }
        with create_temp_config_file(cfg) as f:
            decider = setup_decider(f.name)

            choice = decider.get_bool("dc_bool", self.ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.val(), bool_val)

    def test_get_bool_bad_init(self):
        decider = rust_decider.init("foo", "bah")
        dc = decider.get_bool("bar", self.ctx)

        self.assertEqual(dc.err(), "Decider not found.")

    def test_get_bool_feature_not_found(self):
        with self.test_feature_not_found() as d:
            result = d.get_bool("any", self.ctx)

        self.assertEqual(result.err(), 'Feature "any" not found')

    def test_get_int(self):
        int_val = 99
        cfg = {
            "dc_int": {
                "id": 4393,
                "value": int_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_int",
                "value_type": "Integer",
                "experiment": {"experiment_version": 3},
            }
        }
        with create_temp_config_file(cfg) as f:
            decider = setup_decider(f.name)

            choice = decider.get_int("dc_int", self.ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.val(), int_val)

    def test_get_int_bad_init(self):
        decider = rust_decider.init("foo", "bah")
        dc = decider.get_int("bar", self.ctx)

        self.assertEqual(dc.err(), "Decider not found.")

    def test_get_int_feature_not_found(self):
        with self.test_feature_not_found() as d:
            result = d.get_int("any", self.ctx)

        self.assertEqual(result.err(), 'Feature "any" not found')

    def test_get_float(self):
        float_val = 3.2
        cfg = {
            "dc_float": {
                "id": 5393,
                "value": float_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_float",
                "value_type": "Float",
                "experiment": {"experiment_version": 4},
            }
        }
        with create_temp_config_file(cfg) as f:
            decider = setup_decider(f.name)

            choice = decider.get_float("dc_float", self.ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.val(), float_val)

    def test_get_float_bad_init(self):
        decider = rust_decider.init("foo", "bah")
        dc = decider.get_float("bar", self.ctx)

        self.assertEqual(dc.err(), "Decider not found.")

    def test_get_float_feature_not_found(self):
        with self.test_feature_not_found() as d:
            result = d.get_float("any", self.ctx)

        self.assertEqual(result.err(), 'Feature "any" not found')

    def test_get_string(self):
        string_val = "some_string"
        cfg = {
            "dc_string": {
                "id": 6393,
                "value": string_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_string",
                "value_type": "String",
                "experiment": {"experiment_version": 5},
            }
        }
        with create_temp_config_file(cfg) as f:
            decider = setup_decider(f.name)

            choice = decider.get_string("dc_string", self.ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.val(), string_val)

    def test_get_string_bad_init(self):
        decider = rust_decider.init("foo", "bah")
        dc = decider.get_string("bar", self.ctx)

        self.assertEqual(dc.err(), "Decider not found.")

    def test_get_string_feature_not_found(self):
        with self.test_feature_not_found() as d:
            result = d.get_string("any", self.ctx)

        self.assertEqual(result.err(), 'Feature "any" not found')

    def test_get_map(self):
        map_val = {
            "v": {"nested_map": {"w": True, "x": 1, "y": "some_string", "z": 3.0}},
            "w": False,
            "x": 1,
            "y": "some_string",
            "z": 3.0,
        }
        cfg = {
            "dc_map": {
                "id": 1000000,
                "value": map_val,
                "type": "dynamic_config",
                "version": "1",
                "enabled": True,
                "owner": "test",
                "name": "dc_map",
                "value_type": "Map",
                "experiment": {"experiment_version": 1},
            }
        }
        with create_temp_config_file(cfg) as f:
            decider = setup_decider(f.name)

            choice = decider.get_map("dc_map", self.ctx)

            self.assertEqual(choice.err(), None)
            self.assertEqual(choice.val(), map_val)

    def test_get_map_bad_init(self):
        decider = rust_decider.init("foo", "bah")
        dc = decider.get_map("bar", self.ctx)

        self.assertEqual(dc.err(), "Decider not found.")

    def test_get_map_feature_not_found(self):
        with self.test_feature_not_found() as d:
            result = d.get_map("any", self.ctx)

        self.assertEqual(result.err(), 'Feature "any" not found')

    def test_get_all_values(self):
        bool_val = True
        cfg_bool = {
            "dc_bool": {
                "id": 3393,
                "value": bool_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_bool",
                "value_type": "Boolean",
                "experiment": {"experiment_version": 2},
            }
        }
        cfg_missing_bool = {}
        cfg_missing_bool["dc_missing_bool"] = cfg_bool["dc_bool"].copy()
        cfg_missing_bool["dc_missing_bool"].update(
            {"value": None, "name": "dc_missing_bool"}
        )

        int_val = 99
        cfg_int = {
            "dc_int": {
                "id": 4393,
                "value": int_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_int",
                "value_type": "Integer",
                "experiment": {"experiment_version": 3},
            }
        }
        cfg_missing_int = {}
        cfg_missing_int["dc_missing_int"] = cfg_int["dc_int"].copy()
        cfg_missing_int["dc_missing_int"].update(
            {"value": None, "name": "dc_missing_int"}
        )

        float_val = 3.2
        cfg_float = {
            "dc_float": {
                "id": 5393,
                "value": float_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_float",
                "value_type": "Float",
                "experiment": {"experiment_version": 4},
            }
        }
        cfg_missing_float = {}
        cfg_missing_float["dc_missing_float"] = cfg_float["dc_float"].copy()
        cfg_missing_float["dc_missing_float"].update(
            {"value": None, "name": "dc_missing_float"}
        )

        string_val = "some_string"
        cfg_string = {
            "dc_string": {
                "id": 6393,
                "value": string_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_string",
                "value_type": "String",
                "experiment": {"experiment_version": 5},
            }
        }
        cfg_text = {
            "dc_text": {
                "id": 6393,
                "value": string_val,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_text",
                "value_type": "Text",
                "experiment": {"experiment_version": 5},
            }
        }
        cfg_missing_string = {}
        cfg_missing_string["dc_missing_string"] = cfg_string["dc_string"].copy()
        cfg_missing_string["dc_missing_string"].update(
            {"value": None, "name": "dc_missing_string"}
        )

        cfg_missing_text = {}
        cfg_missing_text["dc_missing_text"] = cfg_text["dc_text"].copy()
        cfg_missing_text["dc_missing_text"].update(
            {"value": None, "name": "dc_missing_text"}
        )

        map_val = {
            "v": {"nested_map": {"w": True, "x": 1, "y": "some_string", "z": 3.0}},
            "w": False,
            "x": 1,
            "y": "some_string",
            "z": 3.0,
        }
        cfg_map = {
            "dc_map": {
                "id": 1000000,
                "value": map_val,
                "type": "dynamic_config",
                "version": "1",
                "enabled": True,
                "owner": "test",
                "name": "dc_map",
                "value_type": "Map",
                "experiment": {"experiment_version": 1},
            }
        }
        cfg_missing_map = {}
        cfg_missing_map["dc_missing_map"] = cfg_map["dc_map"].copy()
        cfg_missing_map["dc_missing_map"].update(
            {"value": None, "name": "dc_missing_map"}
        )

        missing_value_type_cfg = {
            "dc_missing_value_type": {
                "id": 3393,
                "value": False,
                "type": "dynamic_config",
                "version": "2",
                "enabled": True,
                "owner": "test",
                "name": "dc_missing_value_type",
                "experiment": {"experiment_version": 2},
            }
        }

        self.genexp_0_cfg.update(cfg_bool)
        self.genexp_0_cfg.update(cfg_int)
        self.genexp_0_cfg.update(cfg_float)
        self.genexp_0_cfg.update(cfg_string)
        self.genexp_0_cfg.update(cfg_text)
        self.genexp_0_cfg.update(cfg_map)

        # should be set to default values
        self.genexp_0_cfg.update(cfg_missing_bool)
        self.genexp_0_cfg.update(cfg_missing_int)
        self.genexp_0_cfg.update(cfg_missing_float)
        self.genexp_0_cfg.update(cfg_missing_string)
        self.genexp_0_cfg.update(cfg_missing_text)
        self.genexp_0_cfg.update(cfg_missing_map)

        # missing "value_type" field
        self.genexp_0_cfg.update(missing_value_type_cfg)

        # should be excluded from result
        self.genexp_0_cfg.update(self.additional_2_exp)

        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider(f.name)

            values_res = decider.get_all_values(self.ctx)

            self.assertEqual(values_res.err(), None)
            decisions_dict = values_res.decisions()

            # 6 correct DCs, 6 DCs w/ values set to respective defaults
            # 1 `missing_value_type_cfg` which sets "type" to empty string
            # (2 regular experiments are excluded)
            self.assertEqual(len(decisions_dict), 13)

            # test values get set
            bool_val_res = decisions_dict["dc_bool"]
            self.assertEqual(bool_val_res.err(), None)
            self.assertEqual(
                bool_val_res.value_dict(),
                {"name": "dc_bool", "value": bool_val, "type": "boolean"},
            )

            int_val_res = decisions_dict["dc_int"]
            self.assertEqual(int_val_res.err(), None)
            self.assertEqual(
                int_val_res.value_dict(),
                {"name": "dc_int", "value": int_val, "type": "integer"},
            )

            float_val_res = decisions_dict["dc_float"]
            self.assertEqual(float_val_res.err(), None)
            self.assertEqual(
                float_val_res.value_dict(),
                {"name": "dc_float", "value": float_val, "type": "float"},
            )

            string_val_res = decisions_dict["dc_string"]
            self.assertEqual(string_val_res.err(), None)
            self.assertEqual(
                string_val_res.value_dict(),
                {"name": "dc_string", "value": string_val, "type": "string"},
            )

            text_val_res = decisions_dict["dc_text"]
            self.assertEqual(text_val_res.err(), None)
            self.assertEqual(
                text_val_res.value_dict(),
                {"name": "dc_text", "value": string_val, "type": "string"},
            )

            map_val_res = decisions_dict["dc_map"]
            self.assertEqual(map_val_res.err(), None)
            self.assertEqual(
                map_val_res.value_dict(),
                {"name": "dc_map", "value": map_val, "type": "map"},
            )

            # test default values
            missing_bool_val_res = decisions_dict["dc_missing_bool"]
            self.assertEqual(missing_bool_val_res.err(), None)
            self.assertEqual(
                missing_bool_val_res.value_dict(),
                {"name": "dc_missing_bool", "value": False, "type": "boolean"},
            )

            missing_int_val_res = decisions_dict["dc_missing_int"]
            self.assertEqual(missing_int_val_res.err(), None)
            self.assertEqual(
                missing_int_val_res.value_dict(),
                {"name": "dc_missing_int", "value": 0, "type": "integer"},
            )

            missing_float_val_res = decisions_dict["dc_missing_float"]
            self.assertEqual(missing_float_val_res.err(), None)
            self.assertEqual(
                missing_float_val_res.value_dict(),
                {"name": "dc_missing_float", "value": 0.0, "type": "float"},
            )

            missing_string_val_res = decisions_dict["dc_missing_string"]
            self.assertEqual(missing_string_val_res.err(), None)
            self.assertEqual(
                missing_string_val_res.value_dict(),
                {"name": "dc_missing_string", "value": "", "type": "string"},
            )

            missing_text_val_res = decisions_dict["dc_missing_text"]
            self.assertEqual(missing_text_val_res.err(), None)
            self.assertEqual(
                missing_text_val_res.value_dict(),
                {"name": "dc_missing_text", "value": "", "type": "string"},
            )

            missing_map_val_res = decisions_dict["dc_missing_map"]
            self.assertEqual(missing_map_val_res.err(), None)
            self.assertEqual(
                missing_map_val_res.value_dict(),
                {"name": "dc_missing_map", "value": {}, "type": "map"},
            )

            # set "type" to empty string if "value_type" is missing on cfg
            missing_map_val_res = decisions_dict["dc_missing_value_type"]
            self.assertEqual(missing_map_val_res.err(), None)
            self.assertEqual(
                missing_map_val_res.value_dict(),
                {"name": "dc_missing_value_type", "value": False, "type": "boolean"},
            )

    def test_get_experiment(self):
        with create_temp_config_file(self.genexp_0_cfg) as f:
            decider = setup_decider(f.name)

            experiment = decider.get_experiment("genexp_0")
            exp_dict = experiment.val()

            cfg = self.genexp_0_cfg["genexp_0"]
            expected_dict = {
                "id": cfg["id"],
                "name": cfg["name"],
                "enabled": cfg["enabled"],
                "type": "range_variant",
                "emit_event": cfg["emit_event"],
                "measured": False,
                "version": cfg["experiment"]["experiment_version"],
                "value": None,
                "value_type": None,
                "targeting": cfg.get("targeting"),
                "overrides": cfg.get("overrides"),
                "variant_set": {
                    "start_ts": cfg["start_ts"],
                    "stop_ts": cfg["stop_ts"],
                    "shuffle_version": cfg["experiment"]["shuffle_version"],
                    "bucket_val": cfg["experiment"]["bucket_val"],
                    "holdout": None,
                    "mutex_group": None,
                },
            }

            # pythonize() doesn't round range_start/range_end floats well
            # (e.g. 0.2 -> 0.20000000298023224)
            # but we don't use "variants" when calling `get_experiment()` in experiments.py SDK
            # so we ignore comparison here
            del exp_dict["variant_set"]["variants"]

            self.assertEqual(experiment.err(), None)
            self.assertDictEqual(exp_dict, expected_dict)

    def test_get_experiment_bad_init(self):
        decider = rust_decider.init("foo", "bah")
        exp = decider.get_experiment("bar")

        self.assertEqual(exp.err(), "Decider not found.")

    def test_get_experiment_feature_not_found(self):
        with self.test_feature_not_found() as d:
            result = d.get_experiment("any")

        self.assertEqual(result.err(), 'Feature "any" not found')
